# -*- coding: utf-8 -*-
#
#       surfacecache.py
#
#       Copyright 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.surfacecache` -- pages prerendering and caching
--------------------------------------------------------------

This modules contains stuff needed for caching pages and prerendering them. This
is done by the :class:`~pympress.surfacecache.SurfaceCache` class, using several
`dict` of :class:`~cairo.ImageSurface` for storing rendered pages.

The problem is, neither Gtk+ nor Poppler are particularly threadsafe.
Hence the prerendering isn't really done in parallel in another thread, but
scheduled on the main thread at idle times using GLib.idle_add().
"""

import logging
logger = logging.getLogger(__name__)

import threading
import functools
import collections

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import GLib


class SurfaceCache(object):
    """ Pages caching and prerendering made (almost) easy.

    Args:
        doc (:class:`~pympress.document.Document`):  the current document
        max_pages (`int`): The maximum page number.
    """

    #: The actual cache. The `dict`s keys are widget names and its values are
    #: :class:`~collections.OrderedDict`, whose keys are page numbers
    #: and values are instances of :class:`~cairo.ImageSurface`.
    #: In each :class:`~collections.OrderedDict` keys are ordered by
    #: Least Recently Used (get or set), when the size is beyond
    #: :attr:`max_pages`, pages are popped from the start of the cache.
    surface_cache = {}

    #: `dict` containing functions that return a :class:`~cairo.Surface` given a :class:`~cairo.Format`,
    #: width `int` and height `int`, see :meth:`~Gtk.Window.create_similar_image_surface`
    surface_factory = {}

    #: Size of the different managed widgets, as a `dict` of tuples
    surface_size = {}

    #: Type of document handled by each widget. It is a `dict`: its keys are
    #: widget names and its values are document types from ui.
    surface_type = {}

    #: Dictionary of :class:`~threading.Lock` used for managing conccurent
    #: accesses to :attr:`surface_cache` and :attr:`surface_size`
    locks = {}

    #: Set of widgets for which we ignore the max
    unlimited = set()

    #: The current :class:`~pympress.document.Document`.
    doc = None

    #: :class:`~threading.Lock` used to manage conccurent accesses to :attr:`doc`.
    doc_lock = None

    #: Set of active widgets
    active_widgets = set()

    #: maximum number fo pages we keep in cache
    max_pages = 200

    def __init__(self, doc, max_pages):
        self.max_pages = max_pages
        self.doc = doc
        self.doc_lock = threading.Lock()


    def add_widget(self, widget, wtype, prerender_enabled = True, zoomed = False, ignore_max = False):
        """ Add a widget to the list of widgets that have to be managed (for caching and prerendering).

        This creates new entries for ``widget_name`` in the needed internal data
        structures, and creates a new thread for prerendering pages for this widget.

        Args:
            widget (:class:`~Gtk.Widget`):  The widget for which we need to cache
            wtype (`int`):  type of document handled by the widget (see :attr:`surface_type`)
            prerender_enabled (`bool`):  whether this widget is initially in the list of widgets to prerender
            zoomed (`bool`): whether we will cache a zoomed portion of the widget
            ignore_max (`bool`): whether we will cache an unlimited number of slides
        """
        widget_name = widget.get_name().rstrip('0123456789') + ('_zoomed' if zoomed else '')
        with self.locks.setdefault(widget_name, threading.Lock()):
            self.surface_cache[widget_name] = collections.OrderedDict()
            self.surface_size[widget_name] = (-1, -1)
            self.surface_type[widget_name] = wtype
            self.surface_factory[widget_name] = functools.partial(self._create_surface, widget)
            if prerender_enabled and not zoomed:
                self.enable_prerender(widget_name)
            if ignore_max:
                self.unlimited.add(widget_name)


    def swap_document(self, new_doc):
        """ Replaces the current document for which to cache slides with a new one.

        This function also clears the cached pages, since they now belong to an outdated document.

        Args:
            new_doc (:class:`~pympress.document.Document`):  the new document
        """
        with self.doc_lock:
            self.doc = new_doc

        self.clear_cache()


    def disable_prerender(self, widget_name):
        """ Remove a widget from the ones to be prerendered.

        Args:
            widget_name (`str`):  string used to identify a widget
        """
        self.active_widgets.discard(widget_name)


    def enable_prerender(self, widget_name):
        """ Add a widget to the ones to be prerendered.

        Args:
            widget_name (`str`):  string used to identify a widget
        """
        self.active_widgets.add(widget_name)


    def set_widget_type(self, widget_name, wtype):
        """ Set the document type of a widget.

        Args:
            widget_name (`str`):  string used to identify a widget
            wtype (`int`):  type of document handled by the widget (see :attr:`surface_type`)
        """
        with self.locks[widget_name]:
            if self.surface_type[widget_name] != wtype:
                self.surface_type[widget_name] = wtype
                self.surface_cache[widget_name].clear()


    def get_widget_type(self, widget_name):
        """ Get the document type of a widget.

        Args:
            widget_name (`str`):  string used to identify a widget

        Returns:
            `int`: type of document handled by the widget (see :attr:`surface_type`)
        """
        return self.surface_type[widget_name]


    def clear_cache(self, widget_name=None):
        """ Remove all cached values for a given widget. Useful for zoomed views.

        Args:
            widget_name (`str`):  name of the widget that is resized, `None` for all widgets.
        """
        for widget in [widget_name] if widget_name is not None else self.locks:
            with self.locks[widget]:
                self.surface_cache[widget].clear()


    def resize_widget(self, widget_name, width, height):
        """ Change the size of a registered widget, thus invalidating all the cached pages.

        Args:
            widget_name (`str`):  name of the widget that is resized
            width (`int`):  new width of the widget
            height (`int`):  new height of the widget
        """
        with self.locks[widget_name]:
            if (width, height) != self.surface_size[widget_name]:
                self.surface_cache[widget_name].clear()
                self.surface_size[widget_name] = (width, height)


    def get(self, widget_name, page_nb):
        """ Fetch a cached, prerendered page for the specified widget.

        Args:
            widget_name (`str`):  name of the concerned widget
            page_nb (`int`):  number of the page to fetch in the cache

        Returns:
            :class:`~cairo.ImageSurface`: the cached page if available, or `None` otherwise
        """
        with self.locks[widget_name]:
            pc = self.surface_cache[widget_name]
            if page_nb in pc:
                pc.move_to_end(page_nb)
                return pc[page_nb]
            else:
                return None


    def set(self, widget_name, page_nb, val):
        """ Store a rendered page in the cache.

        Args:
            widget_name (`str`):  name of the concerned widget
            page_nb (`int`):  number of the page to store in the cache
            val (:class:`~cairo.ImageSurface`):  content to store in the cache
        """
        with self.locks[widget_name]:
            pc = self.surface_cache[widget_name]
            pc[page_nb] = val
            pc.move_to_end(page_nb)

            while len(pc) > self.max_pages:
                pc.popitem(False)


    def _create_surface(self, widget, fmt, width, height):
        """ Given a widget, create a cairo Image surface with appropriate size and scaling.

        Args:
            widget (:class:`~Gtk.Widget`): the widget for which we’re caching data.
            fmt (:class:`~cairo.Format`): the format for the new surface
            width (`int`): width of the new surface
            height (`int`): height of the new surface

        Returns:
            :class:`~cairo.ImageSurface`: a new image surface
        """
        window = widget.get_window()
        scale = window.get_scale_factor()
        try:
            return window.create_similar_image_surface(fmt, width * scale, height * scale, scale)
        except cairo.Error:
            logger.warning('Failed creating a {} cache surface sized {}x{} scale {} for widget {}'
                           .format(fmt, width * scale, height * scale, scale, widget.get_name()), exc_info=True)
            # Warn here but handle error at the call site
            raise


    def prerender(self, page_nb):
        """ Queue a page for prerendering.

        The specified page will be prerendered for all the registered widgets.

        Args:
            page_nb (`int`):  number of the page to be prerendered
        """
        for name in self.active_widgets:
            GLib.idle_add(self.renderer, name, page_nb)


    def renderer(self, widget_name, page_nb):
        """ Rendering function.

        This function is meant to be scheduled on the GLib main loop. When run,
        it will go through the following steps:

        - check if the job's result is not already available in the cache
        - render it in a new :class:`~cairo.ImageSurface` if necessary
        - store it in the cache if it was not added there since the beginning of
          the process and the widget configuration is still valid

        Args:
            widget_name (`str`):  name of the concerned widget
            page_nb (`int`):  number of the page to store in the cache
        """
        with self.locks[widget_name]:
            if page_nb in self.surface_cache[widget_name]:
                # Already in cache
                return GLib.SOURCE_REMOVE
            ww, wh = self.surface_size[widget_name]
            wtype = self.surface_type[widget_name]

        if ww < 0 or wh < 0:
            logger.warning('Widget {} with invalid size {}x{} when rendering'.format(widget_name, ww, wh))
            return GLib.SOURCE_REMOVE

        with self.doc_lock:
            page = self.doc.page(page_nb)
            if page is None:
                return GLib.SOURCE_REMOVE

        # Render to a ImageSurface
        try:
            surface = self.surface_factory[widget_name](cairo.Format.RGB24, ww, wh)
        except AttributeError:
            logger.warning('Widget {} was not mapped when rendering'.format(widget_name), exc_info = True)
            return GLib.SOURCE_REMOVE
        except cairo.Error:
            return GLib.SOURCE_REMOVE

        context = cairo.Context(surface)
        page.render_cairo(context, ww, wh, wtype)
        del context

        # Save if possible and necessary − using PDF page numbering
        page_nb = page.page_nb
        with self.locks[widget_name]:
            pc = self.surface_cache[widget_name]
            if (ww, wh) == self.surface_size[widget_name] and page_nb not in pc:
                pc[page_nb] = surface

            if widget_name not in self.unlimited:
                pc.move_to_end(page_nb)
                while len(pc) > self.max_pages:
                    pc.popitem(False)

        return GLib.SOURCE_REMOVE


##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:



# -*- coding: utf-8 -*-
#
#       media_overlays/gst.py
#
#       Copyright 2018 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
:mod:`pympress.media_overlays.gst` -- widget to play videos using Gstreamer's Gst
---------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gst', '1.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gst


from pympress import util
from pympress.media_overlays import base


class GstOverlay(base.VideoOverlay):
    """ Simple Gstramer widget.

    Wraps a simple gstreamer playbin.
    """

    #: A :class:`~Gst.Playbin` to be play videos
    playbin = None
    #: A :class:`~Gst.Base.Sink` to display video content
    sink = None

    #: `int` number of milliseconds between updates
    update_freq = 200

    def __init__(self, *args, **kwargs):
        # Create GStreamer playbin
        self.playbin = Gst.ElementFactory.make('playbin', None)
        self.sink = Gst.ElementFactory.make('gtksink', None)
        self.playbin.set_property('video-sink', self.sink)

        super(GstOverlay, self).__init__(*args, **kwargs)

        self.media_overlay.remove(self.movie_zone)
        self.media_overlay.pack_start(self.sink.props.widget, True, True,  0)
        self.media_overlay.reorder_child(self.sink.props.widget, 0)
        self.sink.props.widget.hide()

        # Create bus to get events from GStreamer playin
        bus = self.playbin.get_bus()
        bus.add_signal_watch()
        bus.enable_sync_message_emission()
        bus.connect('message::eos', lambda *args: GLib.idle_add(self.handle_end))
        bus.connect('message::error', lambda _, msg: logger.error('{} {}'.format(*msg.parse_error())))
        bus.connect('message::state-changed', self.on_state_changed)
        bus.connect('message::duration-changed', lambda *args: GLib.idle_add(self.do_update_duration))


    def is_playing(self):
        """ Returns whether the media is currently playing (and not paused).

        Returns:
            `bool`: `True` iff the media is playing.
        """
        return self.playbin.get_state(0).state == Gst.State.PLAYING


    def _set_file(self, filepath):
        """ Sets the media file to be played by the widget.

        Args:
            filepath (`pathlib.Path`): The path to the media file path
        """
        self.playbin.set_property('uri', filepath.as_uri())
        self.playbin.set_state(Gst.State.READY)


    def mute(self, value):
        """ Mutes or unmutes the player.

        Args:
            value (`bool`): `True` iff this player should be muted
        """
        flags = self.playbin.get_property('flags')
        # Fall back to the documented value if introspection fails,
        # see https://gstreamer.freedesktop.org/documentation/playback/playsink.html?gi-language=python#GstPlayFlags
        audio_flag = util.introspect_flag_value(type(flags), 'audio', 0x02)
        if value:
            flags = flags & ~audio_flag
        else:
            flags = flags | audio_flag
        self.playbin.set_property('flags', flags)
        return False


    def on_state_changed(self, bus, msg):
        """ Callback triggered by playbin state changes.

        Args:
            bus (:class:`~Gst.Bus`): the bus that we are connected to
            msg (:class:`~Gst.Message`): the "state-changed" message
        """
        if msg.src != self.playbin:
            # ignore the playbin's children
            return
        old, new, pending = msg.parse_state_changed()
        if old == Gst.State.READY and new == Gst.State.PAUSED:
            # the playbin goes from READY (= stopped) to PLAYING (via PAUSED)
            self.on_initial_play()


    def on_initial_play(self):
        """ Set starting position, start scrollbar updates, unhide overlay. """
        # set starting position, if needed
        if self.start_pos:
            self.do_set_time(self.start_pos)
        # ensure the scroll bar is updated
        GLib.idle_add(self.do_update_duration)
        GLib.timeout_add(self.update_freq, self.do_update_time)
        # ensure the overlay is visible (if needed)
        if not self.media_type.startswith('audio'):
            self.sink.props.widget.show()


    def do_update_duration(self, *args):
        """ Transmit the change of file duration to the UI to adjust the scroll bar.
        """
        changed, time_ns = self.playbin.query_duration(Gst.Format.TIME)
        self.update_range(max(0, time_ns) / 1e9)


    def do_update_time(self):
        """ Update the current position in the progress bar.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.timeout_add` convention)
        """
        changed, time_ns = self.playbin.query_position(Gst.Format.TIME)
        time = time_ns / 1e9
        self.update_progress(time)
        if self.last_timestamp <= self.end_pos <= time:
            self.handle_end()
        self.last_timestamp = time
        return self.playbin.get_state(0).state in {Gst.State.PLAYING, Gst.State.PAUSED}


    def do_play(self):
        """ Start playing the media file.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        self.playbin.set_state(Gst.State.PLAYING)

        return False


    def do_play_pause(self):
        """ Toggle pause mode of the media.

        Should run on the main thread to ensure we avoid reentrency problems.

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        self.playbin.set_state(Gst.State.PLAYING if not self.is_playing() else Gst.State.PAUSED)

        return False


    def do_stop(self):
        """ Stops playing in the backend player.
        """
        self.playbin.set_state(Gst.State.NULL)
        self.playbin.set_state(Gst.State.READY)
        self.sink.props.widget.hide()

        return False


    def do_set_time(self, time):
        """ Set the player at time `~time`.

        Should run on the main thread to ensure we avoid reentrency problems.

        Args:
            time (`float`): the timestamp, in s

        Returns:
            `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
        """
        self.last_timestamp = time
        self.playbin.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, time * Gst.SECOND)
        return False


    @classmethod
    def setup_backend(cls, gst_opts = []):
        """ Prepare/check the Gst backend.

        Returns:
            `str`: the version of Gst used by the backend
        """
        Gst.init(gst_opts)

        return Gst.version_string()



# -*- coding: utf-8 -*-
#
#       ui.py
#
#       Copyright 2010 Thomas Jost <thomas.jost@gmail.com>
#       Copyright 2015 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.ui` -- GUI management
------------------------------------

This module contains the whole graphical user interface of pympress, which is
made of two separate windows: the Content window, which displays only the
current page in full size, and the Presenter window, which displays both the
current and the next page, as well as a time counter and a clock.

Both windows are managed by the :class:`~pympress.ui.UI` class.
"""

import logging
logger = logging.getLogger(__name__)

import pathlib
import math
import sys
import gc
from urllib.request import url2pathname
from urllib.parse import urlsplit

import gi
import cairo
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk, Gdk, GLib, GdkPixbuf, Gio


from pympress import (
    document, surfacecache, util, pointer, scribble, deck, builder, talk_time, dialog, extras, editable_label
)


class UI(builder.Builder):
    """ Pympress GUI management.
    """
    #: The :class:`~pympress.app.Pympress` instance
    app = None
    #: Content window, as a :class:`~Gtk.Window` instance.
    c_win = None
    #: :class:`~Gtk.AspectFrame` for the Content window.
    c_frame = None
    #: :class:`~Gtk.DrawingArea` for the Content window.
    c_da = None

    #: Presenter window, as a :class:`~Gtk.Window` instance.
    p_win = None
    #: :class:`~Gtk.Box` for the Presenter window.
    p_central = None
    #: :class:`~Gtk.AspectFrame` for the current slide in the Presenter window.
    p_frame_notes = None
    #: :class:`~Gtk.DrawingArea` for the current slide in the Presenter window.
    p_da_notes = None

    #: :class:`~Gtk.Grid` for the next slide(s) in the Presenter window.
    grid_next = None
    #: `list` of :class:`~Gtk.AspectFrame` for the next slide in the Presenter window.
    p_frames_next = None
    #: :class:`~Gtk.DrawingArea` for the next slide in the Presenter window.
    p_das_next = None
    #: `int` the number of next slides currently on display in the “Next slides” pane, initialized to the maximal number
    next_frames_count = 16
    #: `list` of :class:`~Gtk.AspectFrame` for the current slide copy in the Presenter window.
    p_frame_cur = None
    #: :class:`~Gtk.DrawingArea` for the current slide copy in the Presenter window.
    p_da_cur = None

    #: :class:`~Gtk.Frame` for the annotations in the Presenter window.
    p_frame_annot = None

    #: Indicates whether we should delay redraws on some drawing areas to fluidify resizing gtk.paned
    resize_panes = False
    #: Tracks return values of GLib.timeout_add to cancel gtk.paned's redraw callbacks
    redraw_timeout = 0

    #: Whether to use notes mode or not
    notes_mode = document.PdfPage.NONE
    #: Current choice of mode to toggle notes
    chosen_notes_mode = document.PdfPage.RIGHT

    #: Whether to display annotations or not
    show_annotations = True
    #: Whether to display big buttons or not
    show_bigbuttons = True
    #: :class:`~Gtk.ToolButton` big button for touch screens, go to previous slide
    prev_button = None
    #: :class:`~Gtk.ToolButton` big button for touch screens, go to next slide
    next_button = None
    #: :class:`~Gtk.ToolButton` big button for touch screens, go toggle the pointer
    laser_button = None
    #: :class:`~Gtk.ToolButton` big button for touch screens, go to scribble on screen
    highlight_button = None

    #: number of page currently displayed in Content window's miniatures
    current_page = -1
    #: number of page currently displayed in Presenter window's miniatures
    preview_page = -1

    #: track whether we blank the screen
    blanked = False

    #: Dictionary of :class:`~Gtk.Widget` from the presenter window that can be dynamically rearranged
    placeable_widgets = {}
    #: Map of :class:`~Gtk.Paned` to the relative position (`float` between 0 and 1) of its handle
    pane_handle_pos = {}

    #: :class:`~pympress.config.Config` to remember preferences
    config = None

    #: :class:`~pympress.surfacecache.SurfaceCache` instance.
    cache = None

    #: Current :class:`~pympress.document.Document` instance.
    doc = document.EmptyDocument()

    #: Class :class:`~pympress.scribble.Scribble` managing drawing by the user on top of the current slide.
    scribbler = None
    #: Class :class:`~pympress.deck.Overview` displaying a view of all slides
    deck = None
    #: Class :class:`~pympress.extras.Annotations` managing the display of annotations
    annotations = None
    #: Class :class:`~pympress.extras.Media` managing keeping track of and callbacks on media overlays
    medias = None
    #: Class :class:`~pympress.extras.Zoom` managing the zoom level of the current slide.
    zoom = None

    #: Software-implemented laser pointer, :class:`~pympress.pointer.Pointer`
    laser = None

    #: :class:`~pympress.editable_label.PageNumber` displaying and setting current page numbers
    page_number = None

    #: :class:`~pympress.editable_label.EstimatedTalkTime` to set estimated/remaining talk time
    est_time = None
    #: :class:`~pympress.dialog.TimingReport` popup to show how much time was spent on which part
    timing = None
    #: :class:`~pympress.talk_time.TimeCounter` clock tracking talk time (elapsed, and remaining)
    talk_time = None

    #: A :class:`~Gtk.ShortcutsWindow` to show the shortcuts
    shortcuts_window = None
    #: :class:`~pympress.dialog.LayoutEditor` popup to configure the layouts of the presenter window
    layout_editor = None
    #: :class:`~pympress.dialog.AutoPlay` popup to configure automatic playing
    autoplay = None

    #: A :class:`~Gtk.AccelGroup` to store the shortcuts
    accel_group = None
    #: A :class:`~Gio.Menu` to display the recent files to open
    recent_menu = None

    #: A :class:`~pympress.extras.FileWatcher` object to reload modified files
    file_watcher = None

    #: `int` or `None`, may keep track of the Gtk.Application inhibit request
    inhibit_cookie = None

    ##############################################################################
    #############################      UI setup      #############################
    ##############################################################################

    def __init__(self, app, config):
        super(UI, self).__init__()
        self.app = app
        self.config = config

        self.blanked = self.config.getboolean('content', 'start_blanked')

        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            util.load_style_provider(Gtk.CssProvider()),
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        # We may want some additional CSS changes
        self.css_provider = Gtk.CssProvider()
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            self.css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

        self.show_annotations = self.config.getboolean('presenter', 'show_annotations')
        self.chosen_notes_mode = document.PdfPage[self.config.get('notes position', 'horizontal').upper()]
        self.show_bigbuttons = self.config.getboolean('presenter', 'show_bigbuttons')

        # Surface cache
        self.cache = surfacecache.SurfaceCache(self.doc, self.config.getint('cache', 'maxpages'))

        # Make and populate windows
        self.load_ui('presenter')
        self.load_ui('content')
        self.app.add_window(self.p_win)
        self.app.add_window(self.c_win)

        self.load_ui('menu_bar', ext='.xml')
        self.app.set_menubar(self.get_object('menu_bar'))
        self.recent_menu = self.get_object('recent_menu')

        c_full = self.config.getboolean('content', 'start_fullscreen')
        p_full = self.config.getboolean('presenter', 'start_fullscreen')
        self.setup_actions({
            'quit':                  dict(activate=self.app.quit),
            'about':                 dict(activate=self.menu_about),
            'big-buttons':           dict(activate=self.switch_bigbuttons, state=self.show_bigbuttons),
            'show-shortcuts':        dict(activate=self.show_shortcuts),
            'content-fullscreen':    dict(activate=self.switch_fullscreen, state=c_full),
            'presenter-fullscreen':  dict(activate=self.switch_fullscreen, state=p_full),
            'swap-screens':          dict(activate=self.swap_screens),
            'blank-screen':          dict(activate=self.switch_blanked, state=self.blanked),
            'notes-mode':            dict(activate=self.switch_mode, state=False),
            'notes-pos':             dict(activate=self.change_notes_pos, parameter_type=str,
                                          state=self.chosen_notes_mode.name.lower()),
            'annotations':           dict(activate=self.switch_annotations, state=self.show_annotations),
            'validate-input':        dict(activate=self.validate_current_input),
            'cancel-input':          dict(activate=self.cancel_current_input),
            'align-content':         dict(activate=self.adjust_frame_position),
            'next-frames':           dict(activate=self.reconfigure_next_frames, parameter_type=int,
                                          state=self.next_frames_count),
        })

        self.setup_actions({
            'open-file':         dict(activate=self.open_file, parameter_type=str),
            'close-file':        dict(activate=self.close_file),
            'save-file':         dict(activate=self.save_file),
            'save-file-as':      dict(activate=self.save_file_as),
            'pick-file':         dict(activate=self.pick_file),
            'list-recent-files': dict(change_state=self.populate_recent_menu, state=False),
            'page':              dict(activate=lambda gaction, param: self.goto_page(param.get_int64()),
                                      parameter_type=int, state=self.current_page),
            'next-page':         dict(activate=self.doc_goto_next),
            'next-label':        dict(activate=self.doc_label_next),
            'prev-page':         dict(activate=self.doc_goto_prev),
            'prev-label':        dict(activate=self.doc_label_prev),
            'hist-back':         dict(activate=self.doc_hist_prev),
            'hist-forward':      dict(activate=self.doc_hist_next),
            'first-page':        dict(activate=self.doc_goto_home),
            'last-page':         dict(activate=self.doc_goto_end),
        })

        self.zoom = extras.Zoom(self)
        self.scribbler = scribble.Scribbler(self.config, self, self.notes_mode)
        self.deck = deck.Overview(self.config, self, self.notes_mode)
        self.annotations = extras.Annotations(self)
        self.medias = extras.Media(self, self.config)
        self.laser = pointer.Pointer(self.config, self)
        self.est_time = editable_label.EstimatedTalkTime(self)
        self.page_number = editable_label.PageNumber(self, self.config.getboolean('presenter', 'scroll_number'))
        self.timing = dialog.TimingReport(self)
        self.autoplay = dialog.AutoPlay(self)
        self.talk_time = talk_time.TimeCounter(self, self.est_time, self.timing, self.autoplay)
        self.layout_editor = dialog.LayoutEditor(self, self.config)
        self.file_watcher = extras.FileWatcher()
        self.config.register_actions(self)

        # Get placeable widgets. NB, get the highlight one manually from the scribbler class
        self.placeable_widgets = {
            name: self.get_object(widget_name) for name, widget_name in self.config.placeable_widgets.items()
        }
        self.placeable_widgets['highlight'] = self.scribbler.scribble_overlay
        self.placeable_widgets['deck'] = self.deck.deck_viewport

        # Initialize windows
        self.make_cwin()
        self.make_pwin()

        self.connect_signals(self)

        for action, shortcut_list in self.config.shortcuts.items():
            if action == 'highlight-hold-to-erase':
                continue  # Not really an action but we define a shortcut for it
            self.app.set_accels_for_action('app.' + action, shortcut_list)

        # Common to both windows
        self.load_icons()

        # Adjust default visibility of items
        self.prev_button.set_no_show_all(True)
        self.next_button.set_no_show_all(True)
        self.laser_button.set_no_show_all(True)
        self.highlight_button.set_no_show_all(True)
        self.p_frame_annot.set_no_show_all(True)

        self.prev_button.set_visible(self.show_bigbuttons)
        self.next_button.set_visible(self.show_bigbuttons)
        self.laser_button.set_visible(self.show_bigbuttons)
        self.highlight_button.set_visible(self.show_bigbuttons)
        self.p_frame_annot.set_visible(self.show_annotations)
        self.laser.activate_pointermode()

        # Setup screens and show all windows
        self.p_win.parse_geometry(self.config.get('presenter', 'geometry'))
        self.c_win.parse_geometry(self.config.get('content', 'geometry'))

        screen = self.p_win.get_screen()
        self.setup_screens(screen, 'at startup')
        screen.connect('monitors-changed', self.screens_changed)

        self.c_win.show_all()
        self.p_win.show_all()

        # Queue some redraws
        self.c_da.queue_draw()
        self.redraw_panes()
        self.do_page_change(unpause=False)


    def load_icons(self):
        """ Set the icon list for both windows.
        """
        try:
            icon_list = [GdkPixbuf.Pixbuf.new_from_file(i) for i in util.list_icons()]
        except Exception:
            logger.exception('Error loading icons')
            return

        self.c_win.set_icon_list(icon_list)
        self.p_win.set_icon_list(icon_list)


    def make_cwin(self):
        """ Initializes the content window.
        """
        self.c_frame.set_property('yalign', self.config.getfloat('content', 'yalign'))
        self.c_frame.set_property('xalign', self.config.getfloat('content', 'xalign'))

        page_type = self.notes_mode.complement()

        self.cache.add_widget(self.c_da, page_type)
        self.cache.add_widget(self.c_da, page_type, zoomed = True)
        self.c_frame.set_property("ratio", self.doc.page(self.current_page).get_aspect_ratio(page_type))

        colourclass = 'white' if self.config.getboolean('content', 'white_blanking') else 'black'
        self.c_da.get_style_context().add_class(colourclass)
        self.c_win.get_style_context().add_class(colourclass)

        self.c_win.insert_action_group('content', self.c_win.get_action_group('win'))
        self.c_win.insert_action_group('win', None)


    def make_pwin(self):
        """ Initializes the presenter window.
        """
        layout = self.config.get_layout('notes' if self.notes_mode else 'plain')
        pane_handles = self.replace_layout(layout, self.p_central, self.placeable_widgets, self.on_pane_event)
        self.pane_handle_pos.update(pane_handles)

        self.p_frames_next = [self.get_object('p_frame_next{}'.format(n)) for n in range(16)]
        self.p_das_next = [self.get_object('p_da_next{}'.format(n)) for n in range(16)]
        self.reconfigure_next_frames(None, GLib.Variant.new_int64(self.config.getint('presenter', 'next_slide_count')))

        slide_type = self.notes_mode.complement()
        self.cache.add_widget(self.p_da_cur, slide_type)
        self.cache.add_widget(self.p_da_cur, slide_type, zoomed = True)
        # A single cache for all next slides
        self.cache.add_widget(self.p_das_next[0], slide_type)
        self.cache.add_widget(self.p_da_notes, self.notes_mode, prerender_enabled = bool(self.notes_mode))
        self.cache.add_widget(self.scribbler.scribble_p_da, slide_type, prerender_enabled = False)
        self.cache.add_widget(self.scribbler.scribble_p_da, slide_type, zoomed = True)
        self.cache.add_widget(self.deck.deck0, slide_type, prerender_enabled = False, ignore_max = True)

        # set default value
        self.page_number.set_last(self.doc.pages_number())

        # Enable dropping files onto the window
        self.p_win.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        self.p_win.drag_dest_add_text_targets()

        self.p_win.insert_action_group('presenter', self.c_win.get_action_group('win'))
        self.p_win.insert_action_group('win', None)


    def reconfigure_next_frames(self, gaction, param):
        """ Callback to set the number of next frames to preview the the “next slides” panel

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the number of slides as a GVariant
        """
        n_frames = param.get_int64()
        if n_frames < 1:
            n_frames = 1
        if n_frames > 16:
            n_frames = 16

        self.reflow_next_frames(n_frames)
        self.next_frames_count = n_frames
        self.app.set_action_state('next-frames', n_frames)
        self.config.set('presenter', 'next_slide_count', str(n_frames))


    def reflow_next_frames(self, n_frames=None):
        """ Set the number of next frames to preview the the “next slides” panel

        Args:
            n_frames (`int`): the number of frames
        """
        for n in range(self.next_frames_count):
            self.grid_next.remove(self.p_frames_next[n])

        for n in range(4):
            self.grid_next.remove_row(n)
            self.grid_next.remove_column(n)

        if n_frames is None:
            n_frames = self.next_frames_count

        ww, wh = self.grid_next.get_allocated_width(), self.grid_next.get_allocated_height()
        rows, cols = self.compute_frame_grid(ww / wh, n_frames)

        for n in range(n_frames):
            self.grid_next.attach(self.p_frames_next[n], n % cols, n // cols, 1, 1)


    def compute_frame_grid(self, grid_ar, n_frames):
        """ Determine the arragement of frames in a grid to maximise their size given respective aspect ratios

        Args:
            grid_ar (`float`): aspect ratio of grid containing the slides
            n_frames (`int`): the number of frames
        """
        if not n_frames:
            return 0, 0

        page_ratio = self.doc.page(self.preview_page).get_aspect_ratio(self.notes_mode.complement())
        grid_ar /= page_ratio

        # works best when n_frames = rows * cols * grid_ar
        rows = math.sqrt(n_frames / grid_ar)

        # If rows < 1 always ceil
        if rows < 1:
            return 1, n_frames

        # Check which of ceiling or flooring rows is better
        rows = math.floor(rows)
        f_cols, c_cols = (n_frames + rows - 1) // rows, (n_frames + rows) // (rows + 1)

        if min(f_cols, grid_ar * rows) > min(c_cols, grid_ar * (rows + 1)):
            return rows, f_cols
        else:
            return rows + 1, c_cols


    def screens_changed(self, screen):
        """ Handle ::monitors-changed events

        Args:
            screen (:class:`~Gdk.Screen`):  the screen
        """
        self.setup_screens(screen, 'at monitors-changed event')
        GLib.timeout_add(1000, self.setup_screens, screen, 'delayed after monitors-changed event')


    def setup_screens(self, screen, event_name=''):
        """ If multiple monitors, fullscreen windows on monitors according to config.

        Args:
            screen (:class:`~Gdk.Screen`):  the screen
            event_name (`str`):  a description of what caused the screen setup event, for debugging
        """
        display = Gdk.Display.get_default()
        c_monitor, p_monitor, primary, non_primary = ScreenArea.lookup_monitors(display, self.c_win, self.p_win)

        # Noisy, but just for debug level
        logger.debug('Detecting monitor layout ' + event_name + ': ' + ', '.join([
            'Content window is at monitor ' + c_monitor.id,
            'Presenter window is at monitor ' + p_monitor.id,
            'Primary monitor is (or overlapped with) ' + primary.id,
            'No non-primary monitor available' if primary.equal(non_primary)
            else 'A non-primary monitor is (or overlapped with) ' + non_primary.id
        ]))

        c_full = self.config.getboolean('content', 'start_fullscreen')
        p_full = self.config.getboolean('presenter', 'start_fullscreen')

        if not c_full and not p_full:
            return

        if primary.equal(non_primary):
            logger.warning(_('Not starting content or presenter window full screen ' +
                             'because there is only one monitor'))

            c_full, p_full = False, False

        else:
            # To start fullscreen, we need to ensure windows are on individual monitors
            if not c_monitor.equal(p_monitor):
                pass
            elif c_monitor.equal(primary):
                self.move_window(self.c_win, c_monitor, non_primary)
            else:  # not p_monitor.equal(primary)
                self.move_window(self.p_win, p_monitor, primary)

        if p_full:
            self.p_win.fullscreen()
        else:
            self.p_win.unfullscreen()

        if c_full:
            self.c_win.fullscreen()
            self.set_screensaver(True)
        else:
            self.c_win.unfullscreen()
            self.set_screensaver(False)

        self.app.set_action_state('content-fullscreen', c_full)
        self.app.set_action_state('presenter-fullscreen', p_full)


    def move_window(self, win, from_bounds, to_bounds):
        """ Move window from monitor number from_monitor to monitor to_monitor.
        """
        x, y, w, h = win.get_position() + win.get_size()
        win_state = win.get_window().get_state() if win.get_window() is not None else 0

        logger.debug('Moving {} window from screen {} to screen {}'.format(
            'Content' if win is self.c_win else 'Presenter', from_bounds.id, to_bounds.id
        ))

        if (win_state & Gdk.WindowState.FULLSCREEN) != 0:
            win.unfullscreen()
        if (win_state & Gdk.WindowState.MAXIMIZED) != 0:
            win.unmaximize()

        to_w = min(w, to_bounds.width)
        to_h = min(h, to_bounds.height)

        # Get fraction of free space that is left or top of window
        x = (max(0, x - from_bounds.x) / (from_bounds.width - w)) if w < from_bounds.width else 0
        y = (max(0, y - from_bounds.y) / (from_bounds.height - h)) if h < from_bounds.height else 0

        win.resize(to_w, to_h)
        win.move(to_bounds.x + x * (to_bounds.width - to_w), to_bounds.y + y * (to_bounds.height - to_h))

        if (win_state & Gdk.WindowState.MAXIMIZED) != 0:
            win.maximize()
        if (win_state & Gdk.WindowState.FULLSCREEN) != 0:
            win.fullscreen()


    def show_shortcuts(self, *args):
        """ Display the shortcuts window.
        """
        # Use a different builder to load and be able to release the shortcuts window
        shortcuts_builder = builder.Builder()
        shortcuts_builder.load_ui('shortcuts')
        self.shortcuts_window = shortcuts_builder.get_object('shortcuts_window')

        for command, shortcut_list in self.config.items('shortcuts'):
            display_shortcut = shortcuts_builder.get_object('shortcut_' + command.replace('-', '_').replace('::', '_'))
            if display_shortcut is not None:
                display_shortcut.props.accelerator = shortcut_list

        self.shortcuts_window.set_transient_for(self.p_win)
        self.shortcuts_window.show_all()
        self.shortcuts_window.present()


    def close_shortcuts(self, *args):
        """ Destroy the shortcuts window once it is hidden.
        """
        self.shortcuts_window.destroy()
        self.shortcuts_window = None


    ##############################################################################
    ############################   Dynamic resize   ##############################
    ##############################################################################

    def on_configure_da(self, widget, event):
        """ Manage "configure" events for all drawing areas, e.g. resizes.

        We tell the local :class:`~pympress.surfacecache.SurfaceCache` cache about it, so that it can
        invalidate its internal cache for the specified widget and pre-render next pages at a correct size.

        Warning: Some not-explicitly sent signals contain wrong values! Just don't resize in that case,
        since these always seem to happen after a correct signal that was sent explicitly.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has been resized
            event (:class:`~Gdk.Event`):  the GTK event, which contains the new dimensions of the widget
        """
        # Don't trust those
        if not event.send_event:
            return

        self.cache.resize_widget(widget.get_name().rstrip('0123456789'), event.width, event.height)

        if widget is self.c_da:
            self.medias.resize('content')
            self.scribbler.reset_scribble_cache()
            self.scribbler.prerender()
        elif widget is self.p_da_cur:
            self.medias.resize('presenter')


    def on_configure_win(self, widget, event):
        """ Manage "configure" events for both window widgets.

        Args:
            widget (:class:`~Gtk.Widget`):  the window which has been moved or resized
            event (:class:`~Gdk.Event`):  the GTK event, which contains the new dimensions of the widget
        """
        geom = '{}x{}{:+}{:+}'.format(*widget.get_size(), *widget.get_position())

        if widget is self.p_win:
            self.config.set('presenter', 'geometry', geom)
            cw = self.p_central.get_allocated_width()
            ch = self.p_central.get_allocated_height()
            self.scribbler.scribble_off_render.set_size_request(cw, ch)
            self.deck.deck_off_render.set_size_request(cw, ch)

            if self.deck.deck_mode:
                self.deck.reset_grid()

            self.adjust_bottom_bar_font()

        elif widget is self.c_win:
            self.config.set('content', 'geometry', geom)


    def adjust_bottom_bar_font(self):
        """ Scale baseline font size of bottom bar, clipped to 6px..13px. Fonts are then scaled by CSS em indications.
        """
        ww, wh = self.p_win.get_size()
        font_size = max(6, min(13, ww / 120 if self.show_bigbuttons else ww / 75))
        self.css_provider.load_from_data('#bottom {{ font-size: {:.1f}px; }}'.format(font_size).encode())


    def redraw_panes(self):
        """ Handler for :class:`~Gtk.Paned`'s resizing signal.

        Used for delayed drawing events of drawing areas inside the panes.

        This is very useful on windows where resizing gets sluggish if we try to redraw while resizing.
        """
        self.resize_panes = False
        self.reflow_next_frames()
        self.p_da_cur.queue_draw()
        for da in self.p_das_next:
            da.queue_draw()
        if self.notes_mode:
            self.p_da_notes.queue_draw()
        if self.redraw_timeout:
            self.redraw_timeout = 0

        self.config.update_layout_from_widgets(self.layout_name(self.notes_mode), self.p_central.get_children()[0],
                                               self.pane_handle_pos)


    def on_pane_event(self, widget, evt):
        """ Signal handler for gtk.paned events.

        This function allows one to delay drawing events when resizing, and to speed up redrawing when
        moving the middle pane is done (which happens at the end of a mouse resize)

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred (ignored)
            evt (:class:`~Gdk.Event`):  the event that occurred
        """
        if type(evt) == Gdk.EventButton and evt.type == Gdk.EventType.BUTTON_RELEASE:
            self.redraw_panes()
        elif type(evt) == GObject.GParamSpec and evt.name == "position":
            self.resize_panes = True
            if self.redraw_timeout:
                GLib.Source.remove(self.redraw_timeout)
            self.redraw_timeout = GLib.timeout_add(200, self.redraw_panes)


    ############################################################################
    ############################  Program lifetime  ############################
    ############################################################################

    def cleanup(self, *args):
        """ Save configuration and exit the main loop.
        """
        self.scribbler.disable_scribbling()
        self.medias.hide_all()

        self.doc.cleanup_media_files()

        if self.app.get_action_state('content-fullscreen'):
            # In case we used hard-disabling
            self.set_screensaver(disabled=False)


    def menu_about(self, *args):
        """ Display the "About pympress" dialog.

        Handles clicks on the "about" menu.
        """
        about = Gtk.AboutDialog(transient_for = self.p_win)
        pympress = util.get_pympress_meta()
        about.set_program_name('pympress')
        about.set_version(pympress['version'])
        about.set_copyright(_('Contributors:') + '\n' + pympress['contributors'])
        about.set_comments(_('pympress is a little PDF reader written in Python ' +
                             'using Poppler for PDF rendering and GTK for the GUI.\n') +
                           _('Some preferences are saved in ') + str(self.config.path_to_config()) + '\n' +
                           _('Resources are loaded from ') + str(util.get_locale_dir().parent) + '\n' +
                           _('The log is written to ') + str(util.get_log_path()) + '\n\n' +
                           _('Media support uses {}.').format(self.medias.backend_version) + '\n' +
                           _('Python version {}').format(sys.version))
        about.set_website('https://github.com/Cimbali/pympress')
        try:
            about.set_logo(GdkPixbuf.Pixbuf.new_from_file(util.get_icon_path('pympress.png')))
        except Exception:
            logger.exception(_('Error loading icon for about window'))
        about.run()
        about.destroy()


    ##############################################################################
    ############################ Document manangement ############################
    ##############################################################################

    def swap_document(self, doc_uri, page=0, reloading=False):
        """ Replace the currently open document with a new one.

        The new document is possibly and EmptyDocument if doc_uri is None.
        The state of the ui and cache are updated accordingly.

        Args:
            doc_uri (`str`): the URI to the new document
            page (`int`): the page at which to start the presentation
            reloading (`bool`): whether we are reloading or detecting stuff from the document
        """
        if self.unsaved_changes(reloading):
            # If we keep the unsaved changes and were prompted by automatic file change,
            # stop watching for file changes in the future.
            if reloading:
                self.file_watcher.stop_watching()
            return

        run_gc = self.doc.doc is not None
        try:
            self.doc = document.Document.create(self, doc_uri)

            if not reloading and doc_uri:
                Gtk.RecentManager.get_default().add_item(doc_uri)
                self.file_watcher.watch_file(doc_uri, self.reload_document)

            elif not reloading:
                self.file_watcher.stop_watching()

        except GLib.Error:
            if reloading:
                return

            self.doc = document.Document.create(self, None)
            self.error_opening_file(doc_uri)
            self.file_watcher.stop_watching()

        self.current_page = self.preview_page = self.doc.goto(page)

        # Guess notes mode by default if the document has notes
        if not reloading:
            hpref = self.config.get('notes position', 'horizontal')
            vpref = self.config.get('notes position', 'vertical')
            target_mode = self.doc.guess_notes(hpref, vpref, self.current_page)

            if self.notes_mode != target_mode:
                self.switch_mode('notes-mode', target_mode=target_mode)

            # don't toggle from NONE to NONE
            if target_mode:
                self.app.activate_action('notes-pos', target_mode.name.lower())
        else:
            self.doc.set_notes_pos(self.notes_mode.direction())

        # Some things that need updating
        self.cache.swap_document(self.doc)
        self.page_number.set_last(self.doc.pages_number())
        self.page_number.enable_labels(self.doc.has_labels())
        self.autoplay.set_doc_pages(self.doc.pages_number())
        self.medias.purge_media_overlays()
        self.timing.set_document_metadata(self.doc.get_structure().copy(), self.doc.page_labels[:])

        # A new document, restart at time 0, paused
        if not reloading:
            self.talk_time.pause()
            self.talk_time.reset_timer()
            self.page_number.setup_doc_callbacks(self.doc)
            self.deck.setup_doc_callbacks(self.doc)

        self.do_page_change(unpause=False)

        # Now that all references to the old document have been replaced or removed, manually
        # collect garbage to delete objects and release file handles / close file descriptors
        if run_gc:
            gc.collect(1)


    def reload_document(self):
        """ Reload the current document.
        """
        self.swap_document(self.doc.get_uri(), page=self.current_page, reloading=True)


    def populate_recent_menu(self, gaction, is_opening=None):
        """ Callback for the recent document menu.

        Gets the URI and requests the document swap.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            is_opening (:class:`~GLib.Variant`): a wrapped boolean indicating whether the menu is opening or closing.
        """
        if not is_opening.get_boolean():
            self.recent_menu.remove_all()
            return

        for file in Gtk.RecentManager.get_default().get_items():
            if not file.exists() or not file.get_mime_type() == 'application/pdf':
                continue

            item = Gio.MenuItem.new(file.get_display_name(), 'app.open-file')
            item.set_action_and_target_value('app.open-file', GLib.Variant.new_string(file.get_uri()))
            item.set_icon(file.get_gicon())

            self.recent_menu.append_item(item)

            if self.recent_menu.get_n_items() >= 10:
                break


    def on_drag_drop(self, widget, drag_context, x, y, data, info, time):
        """ Receive the drag-drops (as text only). If a file is dropped, open it.

        Args:
            widget (:class:`~Gtk.Widget`): The widget on which the dragged item was dropped
            drag_context (:class:`~Gdk.DragContext`): Context object of the dragging
            x (`float`): position of the drop
            y (`float`):  position of the drop
            data (:class:`~Gtk.SelectionData`): container for the dropped data
            info (`int`): info on the target
            time (`int`): time of the drop
        """
        uri_parts = urlsplit(data.get_text())
        path = pathlib.Path(url2pathname(uri_parts.path) if uri_parts.scheme else uri_parts.path)

        if path.is_file() and path.suffix.lower() == '.pdf':
            self.swap_document(path.as_uri())


    def unsaved_changes(self, reload=False):
        """ Prompt the user about what to do with changes in the document: save, discard, or cancel action

        Args:
            reload (`bool`): The unsaved changes is prompted by reloading the file

        Returns:
            `bool`: `True` iff we need to cancel the current action
        """
        if not self.doc.has_changes():
            return False

        dialog = Gtk.MessageDialog(title=_('Unsaved changes'), transient_for=self.p_win, type=Gtk.MessageType.WARNING)
        dialog.set_position(Gtk.WindowPosition.CENTER)
        if not reload:
            dialog.set_markup(_('Save changes before closing?'))
            dialog.format_secondary_markup(_('Unsaved changes will be lost'))

            dialog.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.YES)
            dialog.add_button(_('_Discard'), Gtk.ResponseType.NO).set_image(
                Gtk.Button.new_from_icon_name('edit-delete', Gtk.IconSize.BUTTON).get_image()
            )
            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
            dialog.set_default_response(Gtk.ResponseType.YES)
        else:
            dialog.set_markup(_('The open file was modified outside of pympress but you have made unsaved changes.') +
                              '\n<b>' + _('Overwrite changes instead of reloading?') + '</b>')
            dialog.format_secondary_markup(_('Saving changes will overwrite the changed file!') + '\n' +
                                           _('Unsaved changes will be lost.'))

            dialog.add_button(_('Overwrite'), Gtk.ResponseType.YES).set_image(
                Gtk.Button.new_from_icon_name('document-save', Gtk.IconSize.BUTTON).get_image()
            )
            dialog.add_button(_('Reload'), Gtk.ResponseType.NO).set_image(
                Gtk.Button.new_from_icon_name('view-refresh', Gtk.IconSize.BUTTON).get_image()
            )
            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
            dialog.set_default_response(Gtk.ResponseType.YES)

        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.YES:
            self.doc.save_changes()

        return response == Gtk.ResponseType.CANCEL or (reload and Gtk.ResponseType == Gtk.ResponseType.YES)


    def save_file_as(self, *args):
        """ Remove the current document.
        """
        # Use a GTK file dialog to choose file
        dialog = Gtk.FileChooserDialog(title = _('Save as...'), transient_for = self.p_win,
                                       action = Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_SAVE_AS, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_position(Gtk.WindowPosition.CENTER)

        filter = Gtk.FileFilter()
        filter.set_name(_('PDF files'))
        filter.add_mime_type('application/pdf')
        filter.add_pattern('*.pdf')
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All files'))
        filter.add_pattern('*')
        dialog.add_filter(filter)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self.doc.save_changes(dialog.get_uri())

        dialog.destroy()


    def pick_file(self, *args):
        """ Ask the user which file he means to open.
        """
        # Use a GTK file dialog to choose file
        dialog = Gtk.FileChooserDialog(title = _('Open...'), transient_for = self.p_win,
                                       action = Gtk.FileChooserAction.OPEN)
        dialog.add_buttons(Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_position(Gtk.WindowPosition.CENTER)

        filter = Gtk.FileFilter()
        filter.set_name(_('PDF files'))
        filter.add_mime_type('application/pdf')
        filter.add_pattern('*.pdf')
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name(_('All files'))
        filter.add_pattern('*')
        dialog.add_filter(filter)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self.swap_document(dialog.get_uri())

        dialog.destroy()


    def error_opening_file(self, uri):
        """ Remove the current document.

        Args:
            uri (`str`): the URI of the document
        """
        # Check if the path is valid
        uri_parts = urlsplit(uri)
        filename = pathlib.Path(url2pathname(uri_parts.path))
        if uri_parts.scheme == 'file' and not filename.exists():
            msg = _('Could not find the file "{}"').format(filename)
        else:
            msg = _('Error opening the file "{}"').format(uri)
        dialog = Gtk.MessageDialog(transient_for=self.p_win, flags=Gtk.DialogFlags.MODAL,
                                   message_type=Gtk.MessageType.ERROR, message_format=msg)
        dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
        dialog.set_position(Gtk.WindowPosition.CENTER)
        dialog.run()
        dialog.destroy()


    def close_file(self, *args):
        """ Remove the current document.
        """
        self.swap_document(None)


    def save_file(self, *args):
        """ Remove the current document.
        """
        self.doc.save_changes()


    def open_file(self, gaction, target):
        """ Open a document.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target (:class:`~GLib.Variant`): the file to open as a string variant
        """
        self.swap_document(target.get_string())


    def get_notes_mode(self):
        """ Simple getter.

        Returns:
             :class:`~pympress.document.PdfPage`: Truthy when we split slides in content + notes
        """
        return self.notes_mode


    ##############################################################################
    ############################  Displaying content  ############################
    ##############################################################################

    def on_page_change(self, widget, event=None):
        """ Signal handler for current page editing.

        Args:
            widget (:class:`~Gtk.Widget`):  the editable widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.
        """
        widget_text = widget.get_buffer().get_text()
        try:
            display_page_num = int(widget_text)
        except ValueError:
            return
        else:
            self.goto_page(display_page_num - 1)


    def goto_page(self, page, autoplay=False):
        """ Handle going to the page passed as argument

        Args:
            page (`int`): the page to which to go. Will be clipped to document pages.
            autoplay (`bool`): whether this page change was triggered automatically
        """
        self.preview_page = self.doc.goto(page)

        if not self.page_number.editing:
            self.current_page = self.preview_page

        self.do_page_change(autoplay=autoplay)


    def doc_goto_prev(self, gaction=None, param=None):
        """ Handle going to the next page.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        self.goto_page(self.preview_page - 1)


    def doc_goto_next(self, gaction=None, param=None):
        """ Handle going to the previous page.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        if not self.page_number.editing and self.talk_time.paused:
            self.talk_time.unpause()
        else:
            self.goto_page(self.preview_page + 1)


    def doc_label_next(self, gaction=None, param=None):
        """ Handle going to the next page with a different label.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        self.goto_page(self.doc.label_after(self.preview_page))


    def doc_label_prev(self, gaction=None, param=None):
        """ Handle going to the previous page with a different label.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        self.goto_page(self.doc.label_before(self.preview_page))


    def doc_hist_prev(self, gaction=None, param=None):
        """ Handle going to the previous page in the history of visited pages

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        dest = self.doc.hist_prev()
        if dest is not None:
            self.goto_page(dest)


    def doc_hist_next(self, gaction=None, param=None):
        """ Handle going to the next page in the history of visited pages

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        dest = self.doc.hist_next()
        if dest is not None:
            self.goto_page(dest)


    def doc_goto_home(self, gaction=None, param=None):
        """ Handle going to the start of the document

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        self.goto_page(0)


    def doc_goto_end(self, gaction=None, param=None):
        """ Handle going to the end of the document

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        self.goto_page(self.doc.pages_number())


    def do_page_change(self, unpause=True, autoplay=False):
        """ Switch to another page and display it.

        This is a kind of event which is supposed to be called only from the
        :class:`~pympress.document.Document` class.

        Args:
            is_preview (`bool`):  `True` if the page change should not update the content
            unpause (`bool`):  `True` if the page change should unpause the timer, `False` otherwise
            autoplay (`bool`):  `True` if the page change is triggered automatically, otherwise cancel auto play
        """
        is_preview = self.page_number.editing
        if not is_preview:
            self.preview_page = self.current_page
            self.app.set_action_state('page', self.current_page + 1)

        draw_notes = self.notes_mode
        draw_page = draw_notes.complement()

        page_content = self.doc.page(self.current_page)
        page_preview = self.doc.page(self.preview_page)
        pages_next = [self.doc.page(self.preview_page + n + 1) for n in range(self.next_frames_count)]
        page_notes = self.doc.notes_page(self.preview_page)

        # Aspect ratios and queue redraws
        if draw_notes and page_notes is not None:
            note_pr = page_notes.get_aspect_ratio(draw_notes)
            self.p_frame_notes.set_property('ratio', note_pr)
        self.p_da_notes.queue_draw()

        preview_pr = page_preview.get_aspect_ratio(draw_page)

        self.p_frame_cur.set_property('ratio', preview_pr)
        self.p_da_cur.queue_draw()

        if not is_preview:
            content_pr = page_content.get_aspect_ratio(draw_page)
            self.c_frame.set_property('ratio', content_pr)
            self.c_da.queue_draw()

            self.scribbler.scribble_p_frame.set_property('ratio', content_pr)
            self.scribbler.scribble_p_frame.queue_draw()

        next_pr = preview_pr  # A default page ratio
        for page, frame, da in zip(pages_next, self.p_frames_next, self.p_das_next):
            if page is not None:
                next_pr = page.get_aspect_ratio(draw_page)
            frame.set_property('ratio', next_pr)
            da.queue_draw()

        self.annotations.load_annotations(page_preview)

        # Update display -- needs to be different ?
        self.page_number.update_page_numbers(self.preview_page, page_preview.label())

        # Prerender the 4 next pages and the 2 previous ones
        page_max = min(self.doc.pages_number(), self.preview_page + self.next_frames_count + 4)
        page_min = max(0, self.preview_page - 2)
        for p in list(range(self.preview_page + 1, page_max)) + list(range(self.preview_page, page_min, -1)):
            self.cache.prerender(p)

        if is_preview:
            return

        # Remove scribbles and scribbling/zooming modes
        if self.scribbler.page_change_exits:
            self.scribbler.disable_scribbling()
        self.scribbler.page_change(self.preview_page, page_preview.label())
        self.zoom.stop_zooming()

        # Update medias
        self.medias.replace_media_overlays(self.doc.page(self.current_page), draw_page)

        # Start counter if needed
        if unpause:
            self.talk_time.unpause()
        if not autoplay:
            self.autoplay.stop_looping()

        self.timing.transition(self.preview_page, self.talk_time.current_time())


    def on_draw(self, widget, cairo_context):
        """ Manage draw events for both windows.

        This callback may be called either directly on a page change or as an
        event handler by GTK. In both cases, it determines which widget needs to
        be updated, and updates it, using the
        :class:`~pympress.surfacecache.SurfaceCache` if possible.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget to update
            cairo_context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
        """
        if widget is self.c_da:
            # Current page
            if self.blanked:
                return
            page = self.doc.page(self.current_page)
        elif widget is self.p_da_cur or widget is self.scribbler.scribble_p_da:
            # Current page 'preview'
            page = self.doc.page(self.preview_page)
        elif widget is self.p_da_notes:
            # Notes page, aligned with preview
            page = self.doc.notes_page(self.preview_page)
        elif widget in self.p_das_next:
            offset = 1 + int(widget.get_name()[len('p_da_next'):])
            page = self.doc.page(self.preview_page + offset)
            # No next page: just return so we won't draw anything
            if page is None:
                return
        else:
            logger.warning(_('Unknown widget "{}" to draw'.format(widget.get_name())))

        if page is None:
            if widget is not self.p_da_notes or self.notes_mode.direction() != 'page mapping':
                logger.warning(_('Unexpected missing page to draw for widget "{}"'.format(widget.get_name())))
            return

        if not page.can_render():
            return

        name = widget.get_name().rstrip('0123456789')
        nb = page.number()  # Use PDF page numbering for the cache
        wtype = self.cache.get_widget_type(name)
        ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
        window = widget.get_window()
        scale = window.get_scale_factor()

        if self.zoom.scale != 1. and (widget is self.p_da_cur or widget is self.c_da or
                                      widget is self.scribbler.scribble_p_da):
            zoom_matrix = self.zoom.get_matrix(ww, wh)
            name += '_zoomed'
        else:
            zoom_matrix = cairo.Matrix()

        pb = self.cache.get(name, nb)
        if pb is None:
            if self.resize_panes and widget in self.p_das_next + [self.p_da_cur, self.p_da_notes]:
                # too slow to render here when resize_panes things
                return

            # Cache miss: render the page, and save it to the cache
            try:
                pb = window.create_similar_image_surface(cairo.Format.RGB24, ww * scale, wh * scale, scale)
            except cairo.Error:
                logger.warning('Failed creating an RGB24 surface sized {}x{} scale {} for widget {}'
                               .format(ww * scale, wh * scale, scale, name), exc_info=True)
                if widget is self.c_da:
                    cairo_context.save()
                    cairo_context.set_source_rgb(0, 0, 0)
                    cairo_context.fill()
                    cairo_context.paint()
                    cairo_context.restore()
                return

            cairo_prerender = cairo.Context(pb)
            cairo_prerender.transform(zoom_matrix)
            page.render_cairo(cairo_prerender, ww, wh, wtype)

            self.cache.set(name, nb, pb)

            cairo_context.set_source_surface(pb, 0, 0)
            cairo_context.paint()
        else:
            # Cache hit: draw the surface from the cache to the widget
            cairo_context.set_source_surface(pb, 0, 0)
            cairo_context.paint()

        if widget is self.c_da or widget is self.p_da_cur or widget is self.scribbler.scribble_p_da:
            cairo_context.save()
            cairo_context.transform(zoom_matrix)

            self.scribbler.draw_scribble(widget, cairo_context)
            self.zoom.draw_zoom_target(widget, cairo_context)

            cairo_context.restore()

        if widget is self.c_da or widget is self.p_da_cur or widget is self.scribbler.scribble_p_da:
            # do not use the zoom matrix for the pointer, it is relative to the screen not the slide
            self.laser.render_pointer(cairo_context, ww, wh)


    def clear_zoom_cache(self):
        """ Callback to clear the cache of zoomed widgets.
        """
        self.cache.clear_cache(self.c_da.get_name() + '_zoomed')
        self.cache.clear_cache(self.p_da_cur.get_name() + '_zoomed')
        self.cache.clear_cache(self.scribbler.scribble_p_da.get_name() + '_zoomed')


    def redraw_current_slide(self):
        """ Callback to queue a redraw of the current slides (in both winows).
        """
        self.c_da.queue_draw()
        self.p_da_cur.queue_draw()
        self.scribbler.scribble_p_da.queue_draw()


    ##############################################################################
    ############################     User inputs      ############################
    ##############################################################################

    def validate_current_input(self, gaction, param=None):
        """ Handle the action validating the input, if applicable.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        if self.page_number.try_validate():
            return True
        elif self.est_time.try_validate():
            return True

        return False


    def on_key_input(self, widget, event):
        """ Handle key strokes at top level, only for when editing needs to bypass action accelerators

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the key stroke
            event (:class:`~Gdk.Event`):  the GTK event, which contains the key stroke details
        """
        if self.annotations.key_event(widget, event):
            return True
        elif self.scribbler.key_event(widget, event):
            return True

        return False


    def cancel_current_input(self, gaction, param=None):
        """ Handle the action cancelling the input, if applicable.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        if self.page_number.try_cancel():
            return True
        elif self.est_time.try_cancel():
            return True
        elif self.zoom.try_cancel():
            return True
        elif self.scribbler.try_cancel():
            return True
        elif self.deck.try_cancel():
            return True
        elif self.annotations.try_cancel():
            return True
        elif self.autoplay.stop_looping():
            return True

        return False


    def on_scroll(self, widget, event):
        """ Manage scroll events.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred (ignored)
            event (:class:`~Gdk.Event`):  the event that occurred

        Returns:
            `bool`: whether the event was consumed
        """
        if event.type != Gdk.EventType.SCROLL:
            return False

        # send to spinner if it is active
        elif self.page_number.on_scroll(widget, event):
            return True
        else:
            return False


    def track_motions(self, widget, event):
        """ Track mouse motion events.

        Handles mouse motions on the "about" menu.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget that received the mouse motion
            event (:class:`~Gdk.Event`):  the GTK event containing the mouse position

        Returns:
            `bool`: whether the event was consumed
        """
        if self.zoom.track_zoom_target(widget, event):
            return True
        elif self.scribbler.track_scribble(widget, event):
            return True
        elif self.laser.track_pointer(widget, event):
            return True
        else:
            return self.hover_link(widget, event)


    def track_clicks(self, widget, event):
        """ Track mouse press and release events.

        Handles clicks on the slides.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget that received the click
            event (:class:`~Gdk.Event`):  the GTK event containing the click position

        Returns:
            `bool`: whether the event was consumed
        """
        if self.zoom.toggle_zoom_target(widget, event):
            return True
        elif self.scribbler.toggle_scribble(widget, event):
            return True
        elif self.laser.toggle_pointer(widget, event):
            return True
        else:
            return self.click_link(widget, event)


    def click_link(self, widget, event):
        """ Check whether a link was clicked and follow it.

        Handles a click on a slide.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred
            event (:class:`~Gdk.Event`):  the event that occurred

        Returns:
            `bool`: whether the event was consumed
        """
        if event.type == Gdk.EventType.BUTTON_RELEASE:
            return False

        # Where did the event occur?
        if widget in self.p_das_next:
            offset = 1 + int(widget.get_name()[len('p_da_next'):])
            page = self.doc.page(self.preview_page + offset)
        elif widget is self.p_da_notes:
            page = self.doc.notes_page(self.preview_page)
        elif widget is self.p_da_cur:
            page = self.doc.page(self.preview_page)
        else:
            page = self.doc.page(self.current_page)

        if page is None:
            return False

        x, y = self.zoom.get_slide_point(widget, event)
        page_mode = self.notes_mode if widget is self.p_da_notes else self.notes_mode.complement()

        link = page.get_link_at(x, y, page_mode)

        if event.type == Gdk.EventType.BUTTON_PRESS and link is not None:
            link.follow()
            return True
        else:
            return False


    def hover_link(self, widget, event):
        """ Manage events related to hyperlinks, setting the cursor to a pointer if the hovered region is clickable.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred
            event (:class:`~Gdk.Event`):  the event that occurred

        Returns:
            `bool`: whether the event was consumed
        """
        if event.type != Gdk.EventType.MOTION_NOTIFY:
            return False

        # Where did the event occur?
        if widget in self.p_das_next:
            offset = 1 + int(widget.get_name()[len('p_da_next'):])
            page = self.doc.page(self.preview_page + offset)
        elif widget is self.p_da_notes:
            page = self.doc.notes_page(self.preview_page)
        elif widget is self.p_da_cur:
            page = self.doc.page(self.preview_page)
        else:
            page = self.doc.page(self.current_page)

        if page is None:
            return False

        x, y = self.zoom.get_slide_point(widget, event)
        page_mode = self.notes_mode if widget is self.p_da_notes else self.notes_mode.complement()

        if page.get_link_at(x, y, page_mode):
            extras.Cursor.set_cursor(widget, 'pointer')
            return False
        else:
            extras.Cursor.set_cursor(widget, 'parent')
            return True


    def switch_fullscreen(self, gaction, target):
        """ Switch the Content window to fullscreen (if in normal mode) or to normal mode (if fullscreen).

        Screensaver will be disabled when entering fullscreen mode, and enabled
        when leaving fullscreen mode.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget in which the event occurred

        Returns:
            `bool`: whether some window's full screen status got toggled
        """
        if gaction.get_name() == 'content-fullscreen':
            widget = self.c_win
        elif gaction.get_name() == 'presenter-fullscreen':
            widget = self.p_win
        else:
            raise ValueError('Do not know which widget to put full screen')

        if widget != self.c_win and widget != self.p_win:
            logger.error(_("Unknow widget {} to be fullscreened, aborting.").format(widget))
            return False

        toggle_to = not gaction.get_state().get_boolean()
        window = widget.get_window()
        cur_state = (window.get_state() & Gdk.WindowState.FULLSCREEN) if window is not None else False

        if cur_state:
            widget.unfullscreen()
        else:
            widget.fullscreen()

        if gaction.get_name() == 'content-fullscreen':
            self.set_screensaver(disabled=toggle_to)

        gaction.change_state(GLib.Variant.new_boolean(toggle_to))
        return True


    def set_screensaver(self, disabled):
        """ Disable or re-enable the screensaver.

        Args:
            disabled (`bool`): `True` iff the screensaver should be disabled, otherwise enabled.
        """
        if not disabled:
            if self.inhibit_cookie:
                self.app.uninhibit(self.inhibit_cookie)
            elif self.inhibit_cookie is not None:
                util.hard_set_screensaver(disabled=False)
            self.inhibit_cookie = None

        else:
            flags = (Gtk.ApplicationInhibitFlags.LOGOUT | Gtk.ApplicationInhibitFlags.SWITCH |
                     Gtk.ApplicationInhibitFlags.SUSPEND | Gtk.ApplicationInhibitFlags.IDLE)

            self.inhibit_cookie = self.app.inhibit(self.c_win, flags, _("Fullscreen Presentation running"))

            if not self.inhibit_cookie:
                logger.warning(_('Gtk.Application.inhibit failed preventing screensaver, trying hard disabling'))
                util.hard_set_screensaver(disabled=True)


    def update_frame_position(self, widget, user_data):
        """ Callback to preview the frame alignment, called from the Gtk.SpinButton.

        Args:
            widget (:class:`~Gtk.SpinButton`): The button updating the slide alignment in the drawing area widget
            user_data (`str`): The property being set, either the x or y alignment (resp. xalign and yalign).
        """
        self.c_frame.set_property(user_data, widget.get_value())


    def adjust_frame_position(self, *args):
        """ Select how to align the frame on screen.
        """
        win_aspect_ratio = float(self.c_win.get_allocated_width()) / self.c_win.get_allocated_height()

        if win_aspect_ratio <= float(self.c_frame.get_property("ratio")):
            prop = "yalign"
        else:
            prop = "xalign"

        val = self.c_frame.get_property(prop)

        button = Gtk.SpinButton()
        button.set_adjustment(Gtk.Adjustment(lower=0.0, upper=1.0, step_increment=0.01))
        button.set_digits(2)
        button.set_value(val)
        button.connect("value-changed", self.update_frame_position, prop)

        popup = Gtk.Dialog(title = _("Adjust alignment of slides in projector screen"), transient_for = self.p_win)
        popup.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK)

        box = popup.get_content_area()
        box.add(button)
        popup.show_all()
        response = popup.run()
        popup.destroy()

        # revert if we cancelled
        if response == Gtk.ResponseType.CANCEL:
            self.c_frame.set_property(prop, val)
        else:
            self.config.set('content', prop, str(button.get_value()))


    ##############################################################################
    ############################    Option toggles    ############################
    ##############################################################################

    def swap_screens(self, *args):
        """ Swap the monitors on which each window is displayed (if there are 2 monitors at least).
        """
        display = Gdk.Display.get_default()
        c_monitor, p_monitor, primary, non_primary = ScreenArea.lookup_monitors(display, self.c_win, self.p_win)

        if primary.equal(non_primary) == 1 or p_monitor.equal(c_monitor):
            return

        logger.debug('Swapping screens:')
        self.move_window(self.c_win, c_monitor, p_monitor)
        self.move_window(self.p_win, p_monitor, c_monitor)


    def switch_blanked(self, gaction, param):
        """ Switch the blanked mode of the content screen.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None

        Returns:
            `bool`: whether the notes blanking has been toggled.
        """
        self.blanked = not self.blanked
        self.c_da.queue_draw()
        gaction.change_state(GLib.Variant.new_boolean(self.blanked))

        return True


    def layout_name(self, notes_mode):
        """ Return the layout made for the selected notes_mode

        Args:
            notes_mode (:class:`~pympress.document.PdfPage`): the mode/positioning of notes

        Returns:
            `str`: a string representing the appropriate layout
        """
        if self.scribbler.scribbling_mode:
            return 'highlight'
        elif self.deck.deck_mode:
            return 'deck'
        elif notes_mode.direction() == 'page number':
            return 'note_pages'
        elif notes_mode:
            return 'notes'
        else:
            return 'plain'


    def load_layout(self, new):
        """ Replace the current layout

        Args:
            new (`str`): the name of the layout to load, `None` to use current layout automatically
        """
        if new is None:
            new = self.layout_name(self.notes_mode)
        else:
            self.layout_editor.set_current_layout(new)

        pane_handles = self.replace_layout(self.config.get_layout(new), self.p_central,
                                           self.placeable_widgets, self.on_pane_event)
        self.scribbler.adjust_tools_orientation()
        self.pane_handle_pos.update(pane_handles)

        # queue visibility of all newly added widgets, make sure visibility is right
        self.p_central.show_all()
        self.p_frame_annot.set_visible(self.show_annotations)


    def change_notes_pos(self, gaction, target, force=False):
        """ Switch the position of the nodes in the slide.

        Returns:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target (:class:`~GLib.Variant`): the notes position as a string variant
            force (`bool`): Whether to force the notes switch even if it’s already enabled

        Returns:
            `bool`: whether the notes position has been toggled
        """
        target_mode = document.PdfPage[target.get_string().upper()]

        # Redundant toggle, do nothing
        if target_mode == self.chosen_notes_mode:
            return False

        # Update the choice, except for NONE or BEFORE/AFTER or EVEN/ODD
        if target_mode:
            self.chosen_notes_mode = target_mode
            gaction.change_state(target)
            self.config.set('notes position', target_mode.direction(), target_mode.name.lower())

        # Change the notes arrangement if they are enabled or if we are forced to
        if self.notes_mode or force:
            self.switch_mode(self.app.lookup_action('notes-mode'), target_mode=target_mode, force=True)

        return True


    def switch_mode(self, gaction, target_mode=None, force=False):
        """ Switch the display mode to "Notes mode" or "Normal mode" (without notes).

        Returns:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target_mode (:class:`~pympress.document.PdfPage`): the mode to which we should switch
            force (`bool`): Whether to force the mode switch even if it’s already enabled

        Returns:
            `bool`: whether the notes mode has been toggled
        """
        if target_mode is None:
            target_mode = document.PdfPage.NONE if self.notes_mode else self.chosen_notes_mode

        if target_mode == self.notes_mode and not force:
            return False

        self.scribbler.disable_scribbling()
        self.doc.set_notes_pos(target_mode.direction())

        self.load_layout(self.layout_name(target_mode))

        # Transform or clip page number appropriately
        renumber_page = (target_mode.direction() == 'page parity') - (self.notes_mode.direction() == 'page parity')
        if renumber_page < 0:
            self.preview_page *= 2
            self.current_page *= 2
        elif renumber_page > 0:
            self.preview_page //= 2
            self.current_page //= 2
        elif target_mode.direction() == 'page number':
            self.preview_page = min(self.preview_page, self.doc.pages_number())
            self.current_page = min(self.current_page, self.doc.pages_number())

        self.notes_mode = target_mode
        page_type = self.notes_mode.complement()

        self.cache.set_widget_type('c_da', page_type)
        self.cache.set_widget_type('c_da_zoomed', page_type)
        self.cache.set_widget_type('p_da_next', page_type)
        self.cache.set_widget_type('p_da_cur', page_type)
        self.cache.set_widget_type('p_da_cur_zoomed', page_type)
        self.cache.set_widget_type('scribble_p_da', page_type)
        self.cache.set_widget_type('p_da_notes', self.notes_mode)

        if self.notes_mode:
            self.cache.enable_prerender('p_da_notes')
            self.cache.disable_prerender('p_da_cur')
        else:
            self.cache.disable_prerender('p_da_notes')
            self.cache.enable_prerender('p_da_cur')

        self.medias.adjust_margins_for_mode(page_type)
        self.do_page_change(unpause=False)
        self.page_number.set_last(self.doc.pages_number())
        self.app.set_action_state('notes-mode', bool(self.notes_mode))

        return True


    def switch_annotations(self, gaction, target):
        """ Switch the display to show annotations or to hide them.

        Returns:
            gaction (:class:`~Gio.Action`): the action triggering the call
            target (:class:`~GLib.Variant`): the parameter as a variant, or None

        Returns:
            `bool`: whether the mode has been toggled.
        """
        self.show_annotations = not self.show_annotations

        self.p_frame_annot.set_visible(self.show_annotations)
        self.config.set('presenter', 'show_annotations', 'on' if self.show_annotations else 'off')

        if self.show_annotations:
            parent = self.p_frame_annot.get_parent()
            if issubclass(type(parent), Gtk.Paned):
                if parent.get_orientation() == Gtk.Orientation.HORIZONTAL:
                    size = parent.get_allocated_width()
                else:
                    size = parent.get_allocated_height()
                parent.set_position(self.pane_handle_pos[parent] * size)

        GLib.idle_add(self.redraw_panes)

        self.annotations.load_annotations(self.doc.page(self.preview_page))
        gaction.change_state(GLib.Variant.new_boolean(self.show_annotations))

        return True


    def switch_bigbuttons(self, *args):
        """ Toggle the display of big buttons (nice for touch screens).
        """
        self.show_bigbuttons = not self.show_bigbuttons
        if self.show_bigbuttons:
            # potentially reduce font
            self.adjust_bottom_bar_font()

        self.prev_button.set_visible(self.show_bigbuttons)
        self.next_button.set_visible(self.show_bigbuttons)
        self.laser_button.set_visible(self.show_bigbuttons)
        self.highlight_button.set_visible(self.show_bigbuttons)

        if not self.show_bigbuttons:
            # potentially increase font
            self.adjust_bottom_bar_font()

        self.config.set('presenter', 'show_bigbuttons', 'on' if self.show_bigbuttons else 'off')
        self.app.set_action_state('big-buttons', self.show_bigbuttons)


class ScreenArea(object):
    """ Convenience class to represent monitors or windows in terms of the area (position and size) they use on screen
    """
    def most_intersection(self, candidates):
        """ Find the rectangle that intersects most with `~rect` in `~candidates`

        Args:
            candidates (iterable of `ScreenArea`s): The monitor areas to check for intersection

        Returns:
            `ScreenArea`: The best candidate screen area, i.e. that has the largest intersection
        """
        areas = []
        for geom in candidates:
            intersection = geom.intersection(self)
            if intersection is None:
                areas.append(-1)  # Not even 0 for a common bound
            elif intersection.equal(self):
                return geom
            else:
                areas.append(intersection.width * intersection.height)
        else:
            return candidates[areas.index(max(areas))]


    def least_intersection(self, candidates):
        """ Find the rectangle that intersects least with `~rect` in `~candidates`

        Args:
            candidates (iterable of `ScreenArea`s): The monitor areas to check for intersection

        Returns:
            `ScreenArea`: The best candidate screen area, i.e. that has the smallest intersection
        """
        areas = []
        for geom in candidates:
            intersection = self.intersection(geom)
            if intersection is None:
                return geom
            else:
                areas.append(intersection.width * intersection.height)
        else:
            return candidates[areas.index(min(areas))]


    def __init__(self, obj, id=None):
        if isinstance(obj, tuple):
            self.x, self.y, self.width, self.height = obj
        else:
            self.x, self.y, self.width, self.height = obj.x, obj.y, obj.width, obj.height
        if id is not None:
            self.id = id


    def __repr__(self):
        """ Return a complete representation of the object """
        return 'ScreenArea({}at {} size {})'.format(self.id + ' ' if hasattr(self, 'id') else '',
                                                    (self.x, self.y), (self.width, self.height))


    def intersection(self, other):
        """ Compute the intersection of 2 screen areas

        Args:
            other (`ScreenArea`): The screen area to compare with

        Returns:
            `ScreenArea` or `None`: An area representing the intersection, or `None` if there is no intersection
        """
        if self.x + self.width < other.x or self.x > other.x + other.width:
            return None
        if self.y + self.height < other.y or self.y > other.y + other.height:
            return None

        x = max(self.x, other.x)
        w = min(self.x + self.width, other.x + other.width) - x
        y = max(self.y, other.y)
        h = min(self.y + self.height, other.y + other.height) - y
        return ScreenArea((x, y, w, h))


    def equal(self, other):
        """ Check whether 2 areas cover the exact same space

        Args:
            other (`ScreenArea`): The screen area to compare with

        Returns:
            `bool`: `True` iff the areas are identical
        """
        return (self.x, self.y, self.width, self.height) == (other.x, other.y, other.width, other.height)


    def contains(self, other):
        """ Check whether this area contains `~other`

        Args:
            other (`ScreenArea`): The screen area to compare with

        Returns:
            `bool`: `True` iff the area is contained
        """
        intersection = self.intersection(other)
        return intersection is not None and intersection.equal(self)


    def intersects(self, other):
        """ Check whether this area intersects `~other`

        Args:
            other (`ScreenArea`): The screen area to compare with

        Returns:
            `bool`: `True` iff the areas have an intersection
        """
        return self.intersection(other) is None


    @staticmethod
    def lookup_monitors(display, *windows):
        """ Get the info on the monitors

        Args:
            display (:class:`~Gdk.Display`):  the current screen
            *windows (`tuple` of :class:`~Gtk.Window`):  windows for wich to look up the monitor position

        Returns:
            `tuple` of `ScreenArea`: The monitor areas for each window, followed by
                                     the best monitor areas for primary and non-primary monitors
        """
        # Helpful for debugging
        name = ['{} {}'.format(mon.get_manufacturer() or 'Unknown manufacturer', mon.get_model() or 'Unknown model')
                for mon in (display.get_monitor(n) for n in range(display.get_n_monitors()))]

        all_geom = [ScreenArea(display.get_monitor(n).get_geometry(), name[n]) for n in range(display.get_n_monitors())]
        # Remove duplicate monitors (“mirrored”)
        all_geom = [rect for n, rect in enumerate(all_geom) if not any(rect.equal(other) for other in all_geom[:n])]
        # Remove monitors whose area is entirely contained in that of another monitor. NB: union() computes intersection
        all_geom = [rect for n, rect in enumerate(all_geom)
                    if all(not other.contains(rect) for other in all_geom[:n] + all_geom[n + 1:])]

        pos = []
        for win in windows:
            rect = ScreenArea(win.get_position() + win.get_size())
            pos.append(rect.most_intersection(all_geom))

        prim = ScreenArea(display.get_primary_monitor().get_geometry())
        return (*pos, prim.most_intersection(all_geom), prim.least_intersection(all_geom))


##
# Local Variables:
# mode: python
# indent-tabs-mode: nil
# py-indent-offset: 4
# fill-column: 80
# end:



# -*- coding: utf-8 -*-
#
#       extras.py
#
#       Copyright 2021 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.extras` -- Manages the display of fancy extras such as annotations, videos and cursors
-----------------------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import sys
import copy
import itertools

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib

from pympress import builder


class TimingReport(builder.Builder):
    """ Widget tracking and displaying hierachically how much time was spent in each page/section of the presentation.
    """
    #: `list` of time at which each page was reached
    page_time = []
    #: `int` the time at which the clock was reset
    end_time = -1
    #: The :class:`~Gtk.TreeView` containing the timing data to display in the dialog
    timing_treeview = None
    #: A :class:`~Gtk.Dialog` to contain the timing to show
    time_report_dialog = None
    #: `bool` marking whether next page transition should reset the history of page timings
    clear_on_next_transition = False

    #: A `dict` containing the structure of the current document
    doc_structure = {}
    #: A `list` with the page label of each page of the current document
    page_labels = []
    #: `bool` tracking whether a document is opened
    document_open = False

    def __init__(self, parent):
        super(TimingReport, self).__init__()
        self.load_ui('time_report_dialog')
        self.time_report_dialog.set_transient_for(parent.p_win)
        self.time_report_dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)

        self.connect_signals(self)
        parent.setup_actions({
            'timing-report': dict(activate=self.show_report),
        })


    def transition(self, page, time):
        """ Record a transition time between slides.

        Args:
            page (`int`): the page number of the current slide
            time (`int`): the number of seconds elapsed since the beginning of the presentation
        """
        if not self.document_open:
            return

        if self.clear_on_next_transition:
            self.clear_on_next_transition = False
            del self.page_time[:]

        self.page_time.append((page, time))


    def reset(self, reset_time):
        """ A timer reset. Clear the history as soon as we start changing pages again.
        """
        self.end_time = reset_time
        self.clear_on_next_transition = True


    @staticmethod
    def format_time(secs):
        """ Formats a number of seconds as `minutes:seconds`.

        Returns:
            `str`: The formatted time, with 2+ digits for minutes and 2 digits for seconds.
        """
        return '{:02}:{:02}'.format(*divmod(int(secs), 60))


    def set_document_metadata(self, doc_structure, page_labels):
        """ Show the popup with the timing infortmation.

        Args:
            doc_structure (`dict`): the structure of the document
            page_labels (`list`): the page labels for each of the pages
        """
        self.document_open = len(page_labels) != 0

        # Do not update if we only close the document.
        # That way, the report is still accessible when the document is closed.
        if not self.document_open:
            return

        self.doc_structure = doc_structure
        self.page_labels = page_labels

        # Clear the report when there is a new document opened.
        del self.page_time[:]


    def show_report(self, gaction, param=None):
        """ Show the popup with the timing infortmation.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        times = [time for page, time in self.page_time]
        durations = (e - s for s, e in zip(times, times[1:] + [self.end_time]))

        min_time = min(time for page, time in self.page_time) if self.page_time else 0
        infos = {'time': min_time, 'duration': 0, 'children': [], 'page': 0}
        infos['title'] = 'Full presentation'

        for (page, start_time), duration in zip(self.page_time, durations):
            if not duration:
                continue

            infos['duration'] += duration

            # lookup the position of the page in the document structure (section etc)
            lookup = self.doc_structure
            cur_info_pos = infos
            while lookup:
                try:
                    pos = max(p for p in lookup if p <= page)
                except ValueError:
                    break
                item = lookup[pos]
                lookup = item.get('children', None)

                if cur_info_pos['children'] and cur_info_pos['children'][-1]['page'] == pos:
                    cur_info_pos['children'][-1]['duration'] += duration
                else:
                    cur_info_pos['children'].append({'page': pos, 'title': item['title'], 'children': [],
                                                     'duration': duration, 'time': start_time})
                cur_info_pos = cur_info_pos['children'][-1]

            # add the actual page as a leaf node
            label = self.page_labels[page] if 0 <= page < len(self.page_labels) else 'None'
            cur_info_pos['children'].append({'page': page, 'title': _('slide #') + label,
                                             'duration': duration, 'time': start_time})


        treemodel = self.timing_treeview.get_model()
        if treemodel:
            treemodel.clear()

        treemodel = Gtk.TreeStore(str, str, str, str)

        npages = len(self.page_labels)
        maxlen = len(str(npages))

        dfs_info = [(None, infos)]
        while dfs_info:
            first_it, first = dfs_info.pop()
            page = first['page']
            label = self.page_labels[page] if 0 <= page < len(self.page_labels) else 'None'
            label += '\u2007' * (maxlen - len(str(page)))

            last_col = '{} ({}/{})'.format(label, page, npages)
            row = [first['title'], self.format_time(first['time']), self.format_time(first['duration']), last_col]
            it = treemodel.append(first_it, row)

            if 'children' in first:
                dfs_info.extend((it, child) for child in reversed(first['children']))

        self.timing_treeview.set_model(treemodel)
        self.timing_treeview.expand_row(Gtk.TreePath.new_first(), False)

        self.time_report_dialog.run()
        self.time_report_dialog.hide()


class LayoutEditor(builder.Builder):
    """ Widget tracking and displaying hierachically how much time was spent in each page/section of the presentation.
    """
    #: The :class:`~Gtk.TreeView` displaying the hierarchical layouts
    layout_treeview = None
    #: The :class:`~Gtk.TreeModel` containing the model of the layouts to view in the treeview
    layout_treemodel = None
    #: The :class:`~Gtk.ListModel` containing the possible orientations
    orientations_model = None
    #: A :class:`~Gtk.Dialog` to contain the layout edition dialog
    layout_dialog = None
    #: A :class:`~Gtk.Label` to contain the description of the layout
    layout_description = None
    #: A :class:`~Gtk.ComboBoxText` to select the layout to edit
    layout_selector = None
    #: :class:`~pympress.config.Config` to remember preferences
    config = None
    #: :class:`~Gio.Action` containing the number of next frames
    next_frames_action = None
    #: :class:`~Gio.Action` containing the orientation
    hltools_orientation_action = None
    #: `str` containing the layout currently edited
    current_layout = 'plain'
    #: callback, to be connected to :func:`~pympress.ui.UI.load_layout`
    ui_load_layout = lambda *args: None

    layout_descriptions = {
        'notes':      _('Layout for beamer notes on second screen (no current slide preview in notes)'),
        'plain':      _('Plain layout, without note slides'),
        'note_pages': _('Layout for libreoffice notes on separate pages (with current slide preview in notes)'),
        'highlight':  _('Layout to draw on the current slide'),
        'highlight_notes':  _('Layout to draw on the current slide with notes displayed'),
    }

    _model_columns = ['widget', 'has_resizeable', 'resizeable', 'has_orientation', 'orientation', 'next_slide_count',
                      'widget_name']

    def __init__(self, parent, config):
        super(LayoutEditor, self).__init__()
        self.load_ui('layout_dialog')
        self.layout_dialog.set_transient_for(parent.p_win)
        self.layout_dialog.add_button(Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY)
        self.layout_dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.layout_selector.get_child().set_editable(False)

        self.config = config
        self.ui_load_layout = parent.get_callback_handler('load_layout')

        self.connect_signals(self)
        parent.setup_actions({
            'edit-layout': dict(activate=self.show_editor),
        })


    def load_layout(self):
        """ Load the given layout in the treemodel for display and manipulation in the treeview
        """
        self.layout_description.set_text(self.layout_descriptions[self.current_layout])
        self.layout_treemodel.clear()

        # Display names for the widget ids
        names = {
            'box':         _('box'),
            'notes':       _('notes'),
            'current':     _('current slide'),
            'next':        _('next slide(s)'),
            'highlight':   _('highlighting'),
            'annotations': _('annotations (hideable)'),
            'vertical':    _('vertical'),
            'horizontal':  _('horizontal'),
        }

        next_count = self.next_frames_action.get_state().get_int64()
        hltools_orientation = self.hltools_orientation_action.get_state().get_string()
        dfs_info = [(None, self.config.get_layout(self.current_layout))]

        while dfs_info:
            it, node = dfs_info.pop()

            if type(node) is str:
                orientation = names[hltools_orientation] if node == 'highlight' else ''
                next_slides = next_count if node == 'next' else 0
                self.layout_treemodel.append(it, [node, False, None, bool(orientation), orientation, next_slides,
                                                  names[node]])

            else:
                next_it = self.layout_treemodel.append(it, ['box', True, node['resizeable'],
                                                            True, names[node['orientation']], 0, names['box']])
                dfs_info.extend((next_it, child) for child in reversed(node['children']))

        self.layout_treeview.expand_all()


    def set_current_layout(self, layout):
        """ Update which is the layout currently used by the UI

        Args:
            layout (`str`): the layout id
        """
        self.current_layout = layout


    def layout_selected(self, widget, event=None):
        """ Manage events for the layout selector drop-down menu

        Args:
            widget (:class:`~Gtk.ComboBox`):  the widget which has been modified
            event (:class:`~Gdk.Event`):  the GTK event
        """
        self.current_layout = widget.get_active_id()
        self.load_layout()


    def get_info(self, path):
        """ Given a path string, look up the appropriate item in both the actual and GtkStore models

        Args:
            path (`str`):  A string representing a path in the treemodel

        Returns:
            `dict`, :class:`~Gtk.TreeIter`: the node and iterator representing the position in the layout and model
        """
        pos = Gtk.TreePath.new_from_string(path)
        tree_it = self.layout_treemodel.get_iter(pos)
        node = {'children': [self.config.get_layout(self.current_layout)]}
        for n in pos.get_indices():
            node = node['children'][n]
        return node, tree_it


    def resizeable_toggled(self, widget, path):
        """ Handle when box’ resizeable value is toggled

        Args:
            widget (:class:`~Gtk.ComboBox`):  the widget which has been modified
            path (`str`):  A string representing the path to the modfied item
        """
        node, tree_it = self.get_info(path)
        value = not node['resizeable']
        node['resizeable'] = value
        self.layout_treemodel.set_value(tree_it, self._model_columns.index('resizeable'), value)
        self.normalize_layout(reload=False)


    def orientation_changed(self, widget, path, orient_it):
        """ Handle when the orientation of a box is changed

        Args:
            widget (:class:`~Gtk.ComboBox`):  the widget which has been modified
            path (`str`):  A string representing the path to the modfied item
            orient_it (:class:`~Gtk.TreeIter`): the row of the newly selected value in the orientations liststore model
        """
        value = self.orientations_model.get_value(orient_it, 1)
        node, tree_it = self.get_info(path)
        if node == 'highlight':
            self.hltools_orientation_action.activate(GLib.Variant.new_string(value))
        else:
            node['orientation'] = value
        self.layout_treemodel.set_value(tree_it, self._model_columns.index('orientation'), value)
        self.normalize_layout(reload=False)


    def next_slide_count_edited(self, widget, path, value):
        """ Handle when the next slide count is modified

        Args:
            widget (:class:`~Gtk.ComboBox`):  the widget which has been modified
            path (`str`):  A string representing the path to the modfied item
            value (`int`): the new number of next slides
        """
        node, tree_it = self.get_info(path)
        self.layout_treemodel.set_value(tree_it, self._model_columns.index('next_slide_count'), int(value))
        self.next_frames_action.activate(GLib.Variant.new_int64(int(value)))


    def treemodel_to_tree(self, iterator, parent_horizontal=False, parent_resizeable=False):
        """ Recursive function to transform the treemodel back into our dict-based representation of the layout

        Args:
            iterator (:class:`~Gtk.TreeIter`): the position in the treemodel
            parent_horizontal (`bool`): whether the parent node is horizontal
            parent_resieable (`bool`): whether the parent node is resizeable

        Returns:
            `list`: the list of `dict` or `str` representing the widgets at this level
        """
        nodes = []
        while iterator is not None:
            values = self.layout_treemodel.get(iterator, *range(len(self._model_columns[:-2])))
            node = dict(zip(self._model_columns, values))

            # Make the node conform to either a string or a dictionary with 'children' key
            if node.pop('has_resizeable'):
                node['children'] = []
                del node['widget']
            else:
                node = node['widget']

            if self.layout_treemodel.iter_has_child(iterator):
                children = self.treemodel_to_tree(self.layout_treemodel.iter_children(iterator), *(
                    [parent_horizontal, parent_resizeable] if type(node) is str else
                    [node['orientation'] == 'horizontal', node['resizeable']]
                ))
                if len(children) > 1 and type(node) is not str:
                    # Only assign children if there are any, allows to prune empty boxes
                    node['children'] = children
                elif children and type(node) is not str:
                    # Single-child box replaced by its children
                    node = children[0]
                elif children:
                    # Non-box node with children: create a new box and set the non-box as first child
                    node = {'children': [node] + children, 'resizeable': not parent_resizeable,
                            'orientation': 'vertical' if parent_horizontal else 'horizontal'}

            # Only append widgets, and box nodes that have children
            if type(node) is str or node['children']:
                nodes.append(node)

            iterator = self.layout_treemodel.iter_next(iterator)

        return nodes


    def normalize_layout(self, widget=None, drag_context=None, reload=True):
        """ Handler at the end of a drag-and-drop in the treeview

        Here we transform the listmodel modified by drag-and-drop back to a valid `dict` and `str` hierarchy, and then
        trigger the loading of the layout again to display the corrected layout.

        Args:
            widget (:class:`~Gtk.Widget`): The object which received the signal
            drag_context (:class:`~Gdk.DragContext`): the drag context
            reload (`bool`): whether to reload the layout into the treemodel
        """
        layout = self.treemodel_to_tree(self.layout_treemodel.get_iter_first())
        if len(layout) > 1:
            layout = {'children': layout, 'orientation': 'horizontal', 'resizeable': True}
        else:
            layout = layout[0]
        # This validates
        self.config.update_layout_tree(self.current_layout, layout)
        self.ui_load_layout(None)
        if reload:
            self.load_layout()


    def show_editor(self, gaction, param=None):
        """ Show the popup to edit the layout. Gather info to populate it, and handle apply/cancel at the end.

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        restore_layouts = {layout: copy.deepcopy(self.config.get_layout(layout)) for layout in self.layout_descriptions}

        self.next_frames_action = self.get_application().lookup_action('next-frames')
        self.hltools_orientation_action = self.get_application().lookup_action('highlight-tools-orientation')
        restore_next_count = self.next_frames_action.get_state().get_int64()
        restore_hltools_orientation = self.hltools_orientation_action.get_state().get_string()

        self.layout_selector.set_active_id(self.current_layout)
        self.load_layout()

        if self.layout_dialog.run() != Gtk.ResponseType.APPLY:
            for layout_name, layout in restore_layouts.items():
                self.config.update_layout_tree(layout_name, layout)
            self.next_frames_action.activate(GLib.Variant.new_int64(restore_next_count))
            self.hltools_orientation_action.activate(GLib.Variant.new_string(restore_hltools_orientation))
            self.ui_load_layout(None)

        self.layout_dialog.hide()



class AutoPlay(builder.Builder):
    """ Widget and machinery to setup and play slides automatically, optionally in a loop
    """
    #: A :class:`~Gtk.Dialog` to contain the layout edition dialog
    autoplay_dialog = None
    #: The :class:`~Gtk.SpinButton` for the lower page
    autoplay_spin_lower = None
    #: The :class:`~Gtk.SpinButton` for the upper page
    autoplay_spin_upper = None
    #: The :class:`~Gtk.SpinButton` for the transition between slides
    autoplay_spin_time = None
    #: The :class:`~Gtk.CheckButton` to loop
    autoplay_button_loop = None
    #: :class:`~Glib.Source` which is the source id of the periodic slide transition, or `None` if there is no autoplay
    source = None
    #: if the timeout has been paused, `int` which represents the number of milliseconds until the next page slide
    remain = None
    #: callback, to be connected to :func:`~pympress.ui.UI.goto_page`
    goto_page = lambda *args: None

    def __init__(self, parent):
        super(AutoPlay, self).__init__()
        self.load_ui('autoplay')

        self.autoplay_dialog.set_transient_for(parent.p_win)
        self.autoplay_dialog.add_button(Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY)
        self.autoplay_dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.autoplay_dialog.set_default_response(Gtk.ResponseType.APPLY)

        parent.setup_actions({
            'autoplay': dict(activate=self.run),
        })
        self.connect_signals(self)
        self.goto_page = parent.get_callback_handler('goto_page')


    def set_doc_pages(self, n_pages):
        """ Callback for when a document number of pages changes

        Args:
            n_pages (`int`): the number of pages of the loaded document
        """
        self.autoplay_spin_lower.set_range(1, n_pages - 2)
        self.autoplay_spin_lower.set_value(1)
        self.autoplay_spin_upper.set_range(2, n_pages)
        self.autoplay_spin_upper.set_value(n_pages)


    def page_changed(self, spin_button, scroll_direction):
        """ Callback for when a page spin button is modified, maintains a delta of at least 2 pages between first and
        last page of the intended loop. (No loops needed to loop a single slide.)

        Args:
            spin_button (:class:`~Gtk.SpinButton`): The button whose value was changed
            scroll_direction (:class:`~Gtk.ScrollType`): The speed and amount of change
        """
        if spin_button == self.autoplay_spin_lower:
            minval = self.autoplay_spin_lower.get_value() + 2
            if self.autoplay_spin_upper.get_value() < minval:
                self.autoplay_spin_upper.set_value(minval)
        elif spin_button == self.autoplay_spin_upper:
            maxval = self.autoplay_spin_upper.get_value() - 2
            if self.autoplay_spin_lower.get_value() > maxval:
                self.autoplay_spin_lower.set_value(maxval)


    def pause(self):
        """ Pause the looping if it’s running
        """
        if self.source is None or self.remain is not None:
            return
        self.remain = self.source.get_ready_time() - self.source.get_time()
        self.source.set_ready_time(sys.maxsize)


    def unpause(self):
        """ Unpause the looping if it’s paused
        """
        if self.source is None or self.remain is None:
            return
        self.source.set_ready_time(self.source.get_time() + self.remain)
        self.remain = None


    def is_looping(self):
        """ Return whether an auto-playing
        """
        return self.source is not None


    def stop_looping(self):
        """ Stop the auto-playing
        """
        if self.source is not None:
            self.source.destroy()
        self.source = None
        self.remain = None


    def start_looping(self):
        """ Start the auto-playing
        """
        self.stop_looping()

        it = itertools.cycle(range(*self.pages[:2])) if self.pages[2] else iter(range(*self.pages[:2]))
        self.next_page(it)

        self.source = GLib.timeout_source_new(self.pages[3])
        self.source.attach(GLib.MainContext.default())
        self.source.set_callback(self.next_page, it)


    def next_page(self, it):
        """ Callback to turn the page to the next slide

        Args:
            it (`iterator`): An iterator that contains the next pages to load. Stop when there are no more pages.

        Returns:
            `bool`: `True` if he callback needs to be called again, otherwise `False`
        """
        try:
            self.goto_page(next(it), autoplay=True)
        except StopIteration:
            self.stop_looping()
            return False
        else:
            return True


    def get_page_range(self):
        """ Return the autoplay info

        Returns:
            `tuple`: (first page, stop page, looping, delay i ms)
        """
        return self.pages


    def run(self, gaction, param=None):
        """ Show the dialog to setup auto-play, and start the autoplay if « apply » is selected

        Args:
            gaction (:class:`~Gio.Action`): the action triggering the call
            param (:class:`~GLib.Variant`): the parameter as a variant, or None
        """
        reply = self.autoplay_dialog.run()
        self.autoplay_dialog.hide()

        if reply != Gtk.ResponseType.APPLY:
            return

        self.pages = (self.autoplay_spin_lower.get_value_as_int() - 1, self.autoplay_spin_upper.get_value_as_int(),
                      self.autoplay_button_loop.get_active(), int(self.autoplay_spin_time.get_value() * 1000))
        self.start_looping()



# -*- coding: utf-8 -*-
#
#       pointer.py
#
#       Copyright 2017 Cimbali <me@cimba.li>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
"""
:mod:`pympress.pointer` -- Manage when and where to draw a software-emulated laser pointer on screen
----------------------------------------------------------------------------------------------------
"""

import logging
logger = logging.getLogger(__name__)

import enum

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk, GdkPixbuf, GLib

from pympress import util, extras


class PointerMode(enum.Enum):
    """ Possible values for the pointer.
    """
    #: Pointer switched on continuously
    CONTINUOUS = 2
    #: Pointer switched on only manual
    MANUAL = 1
    #: Pointer never switched on
    DISABLED = 0


class Pointer(object):
    """ Manage and draw the software “laser pointer” to point at the slide.

    Displays a pointer of chosen color on the current slide (in both windows), either on all the time or only when
    clicking while ctrl pressed.

    Args:
        config (:class:`~pympress.config.Config`): A config object containing preferences
        builder (:class:`~pympress.builder.Builder`): A builder from which to load widgets
    """
    #: :class:`~GdkPixbuf.Pixbuf` to read XML descriptions of GUIs and load them.
    pointer = GdkPixbuf.Pixbuf()
    #: `(float, float)` of position relative to slide, where the pointer should appear
    pointer_pos = (.5, .5)
    #: `bool` indicating whether we should show the pointer
    show_pointer = False
    #: :class:`~pympress.pointer.PointerMode` indicating the pointer mode
    pointer_mode = PointerMode.MANUAL
    #: The :class:`~pympress.pointer.PointerMode` to which we toggle back
    old_pointer_mode = PointerMode.CONTINUOUS
    #: A reference to the UI's :class:`~pympress.config.Config`, to update the pointer preference
    config = None
    #: :class:`~Gtk.DrawingArea` Slide in the Presenter window, used to reliably set cursors.
    p_da_cur = None
    #: :class:`~Gtk.DrawingArea` Slide in the Contents window, used to reliably set cursors.
    c_da     = None
    #: :class:`~Gtk.AspectFrame` Frame of the Contents window, used to reliably set cursors.
    c_frame  = None
    #: a `dict` of the :class:`~Gtk.RadioMenuItem` selecting the pointer mode
    pointermode_radios = {}

    #: callback, to be connected to :func:`~pympress.ui.UI.redraw_current_slide`
    redraw_current_slide = lambda *args: None
    #: callback, to be connected to :meth:`~pympress.app.Pympress.set_action_state`
    set_action_state = None

    def __init__(self, config, builder):
        super(Pointer, self).__init__()
        self.config = config

        builder.load_widgets(self)

        self.redraw_current_slide = builder.get_callback_handler('redraw_current_slide')
        self.set_action_state = builder.get_callback_handler('app.set_action_state')

        default_mode = config.get('presenter', 'pointer_mode')
        default_color = config.get('presenter', 'pointer')

        try:
            default_mode = PointerMode[default_mode.upper()]
        except KeyError:
            default_mode = PointerMode.MANUAL

        self.activate_pointermode(default_mode)
        self.load_pointer(default_color)

        self.action_map = builder.setup_actions({
            'pointer-color': dict(activate=self.change_pointercolor, state=default_color, parameter_type=str),
            'pointer-mode': dict(activate=self.change_pointermode, state=default_mode.name.lower(), parameter_type=str),
        })


    def load_pointer(self, name):
        """ Perform the change of pointer using its color name.

        Args:
            name (`str`): Name of the pointer to load
        """
        if name not in ['red', 'green', 'blue']:
            raise ValueError('Wrong color name')
        path = util.get_icon_path('pointer_' + name + '.png')
        try:
            self.pointer = GdkPixbuf.Pixbuf.new_from_file(path)
        except Exception:
            logger.exception(_('Failed loading pixbuf for pointer "{}" from: {}'.format(name, path)))


    def change_pointercolor(self, action, target):
        """ Callback for a radio item selection as pointer mode (continuous, manual, none).

        Args:
            action (:class:`~Gio.Action`): The action activatd
            target (:class:`~GLib.Variant`): The selected mode
        """
        color = target.get_string()
        self.load_pointer(color)
        self.config.set('presenter', 'pointer', color)
        action.change_state(target)


    def activate_pointermode(self, mode=None):
        """ Activate the pointer as given by mode.

        Depending on the given mode, shows or hides the laser pointer and the normal mouse pointer.

        Args:
            mode (:class:`~pympress.pointer.PointerMode`): The mode to activate
        """
        # Set internal variables, unless called without mode (from ui, after windows have been mapped)
        if mode == self.pointer_mode:
            return
        elif mode is not None:
            self.old_pointer_mode, self.pointer_mode = self.pointer_mode, mode
            self.config.set('presenter', 'pointer_mode', self.pointer_mode.name.lower())


        # Set mouse pointer and cursors on/off, if windows are already mapped
        self.show_pointer = False
        for slide_widget in [self.p_da_cur, self.c_da]:
            ww, wh = slide_widget.get_allocated_width(), slide_widget.get_allocated_height()
            if max(ww, wh) == 1:
                continue

            window = slide_widget.get_window()
            pointer_coordinates = window.get_pointer() if window is not None else (-1, -1)

            if 0 < pointer_coordinates.x < ww and 0 < pointer_coordinates.y < wh \
                    and self.pointer_mode == PointerMode.CONTINUOUS:
                # Laser activated right away
                self.pointer_pos = (pointer_coordinates.x / ww, pointer_coordinates.y / wh)
                self.show_pointer = True
                extras.Cursor.set_cursor(slide_widget, 'invisible')
            else:
                extras.Cursor.set_cursor(slide_widget, 'parent')

        self.redraw_current_slide()


    def change_pointermode(self, action, target):
        """ Callback for a radio item selection as pointer mode (continuous, manual, none).

        Args:
            action (:class:`~Gio.Action`): The action activatd
            target (:class:`~GLib.Variant`): The selected mode
        """
        if target is None or target.get_string() == 'toggle':
            mode = self.old_pointer_mode if self.pointer_mode == PointerMode.CONTINUOUS else PointerMode.CONTINUOUS
        else:
            mode = PointerMode[target.get_string().upper()]
        self.activate_pointermode(mode)

        action.change_state(GLib.Variant.new_string(mode.name.lower()))


    def render_pointer(self, cairo_context, ww, wh):
        """ Draw the laser pointer on screen.

        Args:
            cairo_context (:class:`~cairo.Context`): The canvas on which to render the pointer
            ww (`int`): The widget width
            wh (`int`): The widget height
        """
        if self.show_pointer:
            x = ww * self.pointer_pos[0] - self.pointer.get_width() / 2
            y = wh * self.pointer_pos[1] - self.pointer.get_height() / 2
            Gdk.cairo_set_source_pixbuf(cairo_context, self.pointer, x, y)
            cairo_context.paint()


    def track_pointer(self, widget, event):
        """ Move the laser pointer at the mouse location.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if self.show_pointer:
            ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
            ex, ey = event.get_coords()
            self.pointer_pos = (ex / ww, ey / wh)
            self.redraw_current_slide()
            return True

        else:
            return False


    def track_enter_leave(self, widget, event):
        """ Switches laser off/on in continuous mode on leave/enter slides.

        In continuous mode, the laser pointer is switched off when the mouse leaves the slide
        (otherwise the laser pointer "sticks" to the edge of the slide).
        It is switched on again when the mouse reenters the slide.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        # Only handle enter/leave events on one of the current slides
        if self.pointer_mode != PointerMode.CONTINUOUS or widget not in [self.c_da, self.p_da_cur]:
            return False

        if event.type == Gdk.EventType.ENTER_NOTIFY:
            self.show_pointer = True
            extras.Cursor.set_cursor(widget, 'invisible')

        elif event.type == Gdk.EventType.LEAVE_NOTIFY:
            self.show_pointer = False
            extras.Cursor.set_cursor(widget, 'parent')

        self.redraw_current_slide()
        return True


    def toggle_pointer(self, widget, event):
        """ Track events defining when the laser is pointing.

        Args:
            widget (:class:`~Gtk.Widget`):  the widget which has received the event.
            event (:class:`~Gdk.Event`):  the GTK event.

        Returns:
            `bool`: whether the event was consumed
        """
        if self.pointer_mode in {PointerMode.DISABLED, PointerMode.CONTINUOUS}:
            return False

        ctrl_pressed = event.get_state() & Gdk.ModifierType.CONTROL_MASK

        if ctrl_pressed and event.type == Gdk.EventType.BUTTON_PRESS:
            self.show_pointer = True
            extras.Cursor.set_cursor(widget, 'invisible')

            # Immediately place & draw the pointer
            return self.track_pointer(widget, event)

        elif self.show_pointer and event.type == Gdk.EventType.BUTTON_RELEASE:
            self.show_pointer = False
            extras.Cursor.set_cursor(widget, 'parent')
            self.redraw_current_slide()
            return True

        else:
            return False
